"""
This module implements gftools/bin/gftools-packager.py

Tool to take files from a font family project upstream git repository
to the google/fonts GitHub repository structure, taking care of all the details.

Documentation at gftools/docs/gftools-packager/README.md
"""
import sys
import os
from pathlib import PurePath
import shutil
from tempfile import TemporaryDirectory, mkstemp
import subprocess
import requests
import pprint
import typing
from collections import OrderedDict
import traceback
from io import StringIO, BytesIO
from contextlib import contextmanager
import urllib.parse
import pygit2 # type: ignore
from strictyaml import ( # type: ignore
                        Map,
                        UniqueSeq,
                        MapPattern,
                        Enum,
                        Str,
                        EmptyNone,
                        EmptyDict,
                        Optional,
                        dirty_load,
                        as_document,
                        YAMLValidationError,
                        YAML
                      )
import functools
from hashlib import sha1
from fontTools.ttLib import TTFont # type: ignore
from gflanguages import LoadLanguages
from gftools.util import google_fonts as fonts
from gftools.github import GitHubClient

# ignore type because mypy error: Module 'google.protobuf' has no
# attribute 'text_format'
from google.protobuf import text_format # type: ignore

# Getting many mypy errors here like: Lib/gftools/fonts_public_pb2.py:253:
#     error: Unexpected keyword argument "serialized_options" for "Descriptor"
# The "type: ignore" annotation didn't help.
from typing import TYPE_CHECKING
if TYPE_CHECKING:
  fonts_pb2: typing.Any
else:
  import gftools.fonts_public_pb2 as fonts_pb2

CATEGORIES = ['DISPLAY', 'SERIF', 'SANS_SERIF', 'HANDWRITING', 'MONOSPACE']

from pkg_resources import resource_filename
with open(resource_filename('gftools', 'template.upstream.yaml')) as f:
  upstream_yaml_template = f.read()
  # string.format fails if we use other instances of {variables}
  # without adding them to the call to format (KeyError).
  upstream_yaml_template = upstream_yaml_template.replace('{CATEGORIES}', ', '.join(CATEGORIES))


GIT_NEW_BRANCH_PREFIX = 'gftools_packager_'
# Using object(expression:$rev), we query all three license folders
# for family_name, but only the entry that exists will return a tree (directory).
# Non existing directories will be null (i.e. None).
# Hence, we can send one query to know the family exists or not, in which
# directory (license) and have a (flat) directory listing.
# I queried "rateLimit{cost}" and this had a cost of 1!
GITHUB_GRAPHQL_GET_FAMILY_ENTRY = """
fragment FamilyFiles on Tree {
  entries{
    name
    type
    oid
  }
}

query ListFiles($repoName: String!,
                $repoOwner: String!,
                $reference: String!,
                $oflDir: String!,
                $uflDir: String!,
                $apacheDir: String!
) {
  repository(name: $repoName, owner: $repoOwner) {
    ref(qualifiedName: $reference) {
      prefix
      name
      target {
        ... on Commit {
          __typename
          oid
          messageHeadline
          pushedDate
        }
      }
    }
    ofl: object(expression: $oflDir) {
      ...FamilyFiles
    }
    apache: object(expression:$apacheDir) {
      __typename
      ...FamilyFiles
    }
    ufl: object(expression:$uflDir) {
      __typename
      ...FamilyFiles
    }
  }
}
"""

def _get_query_variables(repo_owner, repo_name, family_name, reference='refs/heads/main'):
  """
  call like: get_query_variables('google', 'fonts', 'gelasio')

  reference: see $ git help rev-parse
            and git help revisions
            and https://git-scm.com/book/en/v2/Git-Internals-Git-References
  For a branch called "main", "refs/heads/main" is best
  but "main" would work as well.
  Tag names work as well, ideally "ref/tags/v0.6.8" but "v0.6.8" would
  work too. The full name is less ambiguous.
  """
  return {
    'repoOwner': repo_owner,
    'repoName': repo_name,
    'reference': reference,
    'oflDir': f'{reference}:ofl/{family_name}',
    'apacheDir': f'{reference}:apache/{family_name}',
    'uflDir': f'{reference}:ufl/{family_name}'
  }

def _family_name_normal(family_name: str) -> str:
  return family_name.lower()\
      .replace(' ', '')\
      .replace('.', '')\
      .replace('/', '')

def get_gh_gf_family_entry(family_name):
  # needs input sanitation
  family_name_normal = _family_name_normal(family_name)
  variables = _get_query_variables('google','fonts', family_name_normal)

  result = GitHubClient("google", "fonts")._run_graphql(GITHUB_GRAPHQL_GET_FAMILY_ENTRY, variables)
  return result

def _git_tree_iterate(path, tree, topdown):
  dirs = []
  files = []
  for e in tree:
    if e.type == pygit2.GIT_OBJ_TREE:
      dirs.append(e.name)
    elif e.type == pygit2.GIT_OBJ_BLOB:
      files.append(e.name)
  if topdown:
    yield path and os.path.join(*path) or '.', dirs, files
  # note, if topdown, caller can manipulate dirs
  for name in dirs:
    path.append(name)
    yield from _git_tree_iterate(path, tree[name], topdown)
    path.pop()
  if not topdown:
    yield path and os.path.join(*path) or '.', dirs, files

def _git_tree_walk(path, tree, topdown=True):
  yield from _git_tree_iterate(path.split(os.sep), tree[path], topdown)

def get_github_gf_blob(file_sha):
  return GitHubClient('google', 'fonts').get_blob(file_sha)

def _shallow_clone_git(target_dir, git_url, branch_or_tag='main'):
  """
      getting this as a shallow copy, because for some files we want to
      search in the filesystem.

      branch_or_tag: as used in `git clone -b`

      NOTE: libgit2 and hence pygit2 doesn't support shallow clones yet,
      but that's the most lightweight way to get the whole directory
      structure.
  """

  # I don't understand why git clone doesn't take this more explicit form.
  # But, I recommended it in the docs, so here's a little fix.
  if branch_or_tag.startswith('tags/'):
    branch_or_tag = branch_or_tag[len('tags/'):]

  return subprocess.run(['git', 'clone', '--depth', '1', '--bare'
                       , '-b', branch_or_tag, git_url
                       , target_dir], check=True
                       , stdout=subprocess.PIPE)

# Eventually we need all these keys to make an update, so this
# can't have Optional/Empty entries, unless that's really optional for
# the process.
upstream_yaml_schema = Map({
    'name': Str(),
    'repository_url': Str(), # TODO: custom validation please
    'branch': Str(),
    'category': UniqueSeq(Enum(CATEGORIES)),
    'designer': Str(),
    Optional('build', default=''): EmptyNone() | Str(),
    # allowing EmptyDict here, even though we need files in here,
    # but we will catch missing files later in the process.
    # When we have repository_url and branch we can add a editor based
    # dialog that suggests all files present in the repo (delete lines of
    # files you don't want to include).
    'files': EmptyDict() | MapPattern(Str(), Str()) # Mappings with arbitrary key names
})

# Since upstream_yaml_template is incomplete, it can't be parsed with
# the complete upstream_yaml_schema. Here's a more forgiving schema for
# the template and for initializing with a stripped upstream_conf.
upstream_yaml_template_schema = Map({
    Optional('name', default=''): EmptyNone() | Str(),
    Optional('repository_url', default=''): EmptyNone() | Str(), # TODO: custom validation please
    'branch': EmptyNone() | Str(),
    Optional('category', default=None):  EmptyNone() | UniqueSeq(Enum(CATEGORIES)),
    Optional('designer', default=''): EmptyNone() |Str(),
    Optional('build', default=''): EmptyNone() | Str(),
    'files': EmptyDict() | MapPattern(Str(), Str())
})

upstream_yaml_stripped_schema = Map({ # TODO: custom validation please
    # Only optional until it can be in METADATA.pb
    Optional('repository_url', default=''): Str(),
    'branch': EmptyNone() | Str(),
    Optional('build', default=''): EmptyNone() | Str(),
    'files': EmptyDict() | MapPattern(Str(), Str())
})

# ALLOWED FILES
LICENSE_FILES_2_DIRS = (
        ('LICENSE.txt', 'apache')
      , ('UFL.txt', 'ufl')
      , ('OFL.txt', 'ofl')

)

# ('apache', 'ufl', 'ofl')
LICENSE_DIRS = tuple(zip(*LICENSE_FILES_2_DIRS))[1]

# /path/to/google/Fonts$ find */*/*  | grep -E "(ofl|apache|ufl)/*/*" \
#                                    | grep -Ev "*/*/*.ttf" \
#                                    | sed 's/.*\/.*\///g' \
#                                    | sort | uniq -c | sort -hr
#  1094 DESCRIPTION.en_us.html
#  1064 METADATA.pb
#  1034 OFL.txt
#   383 FONTLOG.txt
#    43 LICENSE.txt
#    41 static
#    39 EARLY_ACCESS.category
#     7 README
#     4 COPYRIGHT.txt
#     3 UFL.txt
#     3 TRADEMARKS.txt
#     3 README.txt
#     3 LICENCE.txt
#     3 LICENCE-FAQ.txt
#     3 CONTRIBUTING.txt
#     2 README.md
#     2 CONTRIBUTORS.txt
#     1 ofl.txt
#     1 DESCRIPTION.vi_vn.html
#     1 AUTHORS.txt
#
# NOTE: as seen above, there are a lot more filenames in the family
# directories than we do allow here. This may need some adjustment
# later.
# We could allow them maybe if they are already in the upstream, so new
# packages can't add files we don't want to have anymore. Something like
# this.
ALLOWED_FILES = {
    'DESCRIPTION.en_us.html'
  , 'FONTLOG.txt'
  , *dict(LICENSE_FILES_2_DIRS).keys() # just the file names/keys
# METADATA.pb is not taken from upstream, technically we update the
# version in google fonts or create it newly
}

def _write_file_to_package(basedir:str, filename:str, data:bytes) -> None:
  full_name = os.path.realpath(os.path.join(basedir, filename))

  # Can't just let write the file anywhere!
  full_directory = os.path.join(os.path.realpath(basedir), '')
  if os.path.commonprefix([full_name, full_directory]) != full_directory:
    raise Exception(f'Target is not in package directory: "{filename}".')

  os.makedirs(os.path.dirname(full_name), exist_ok=True)
  with open(full_name, 'wb') as f:
    f.write(data)

def _file_in_package(basedir, filename):
  full_name = os.path.join(basedir, filename)
  return os.path.isfile(full_name)

class UserAbortError(Exception):
  pass

class ProgramAbortError(Exception):
  pass

def _get_gf_dir_content(family_name: str) \
        -> typing.Tuple[typing.Union[str, None], typing.Dict[str, typing.Dict[str, typing.Any]]]:
  gfentry = get_gh_gf_family_entry(family_name)
  entries = None
  for license_dir in LICENSE_DIRS:
    if gfentry['data']['repository'][license_dir] is not None:
      entries = gfentry['data']['repository'][license_dir]['entries']
      break
  if entries is None:
    return None, {}
  gf_dir_content = {f['name']: f for f in entries}
  return license_dir, gf_dir_content


def _get_editor_command():
  # # there's some advice to chose an editor to open and how to set a default
  # https://stackoverflow.com/questions/10725238/opening-default-text-editor-in-bash
  # I like chosing VISUAL over EDITOR falling back to vi, where on my
  # system actually vi equals vim:
  # ${VISUAL:-${EDITOR:-vi}}
  return os.environ.get('VISUAL' , os.environ.get('EDITOR', 'vi'))

# ANSI controls
TOLEFT = '\u001b[1000D' # Move all the way left (max 1000 steps
CLEARLINE = '\u001b[2K'    # Clear the line
# UP =  '\u001b[1A' # moves cursor 1 up
# reset = (CLEARLINE + UP) * num_linebeaks + TOLEFT

def user_input(question: str,
               options: 'OrderedDict[str, str]',
               default: typing.Union[str, None] = None,
               yes: typing.Union[bool, None] = None,
               quiet: bool = False
            ):
  """
    Returns one of the keys of the *options* dict.

    In interactive mode (if *yes* is not True, see below) use the
    *input()* function to ask the user a *question* and present the user
    with the possible answers in *options*. Where the keys in *options*
    are the actual options to enter and the values are the descriptions
    or labels.

    default: if *yes* is a bool this should be an option that does
    not require user interaction. That way we can have an all -y/--no-confirm
    flag will always choose the default.

    yes: don't ask the user and use the default. If the value is a boolean
    *default* must be set, because we expect the boolean comes from the
    -y/--no-confirm flag and the programmers intent is to make this dialogue
    usable with that flag. If the value is None, we don't check if default is
    set. The boolean False versus None differentiation is intended as a self
    check to raise awareness of how to use this function.

    quiet: if *yes* is true don't print the question to stdout.
  """
  if default is not None and default not in options:
    # UX: all possible choices must be explicit.
    raise Exception(f'default is f{default} but must be one of: '
                    f'{", ".join(options.keys())}.')
  if yes is not None and default is None:
    # This is a programming error see the __doc__ string above.
    raise Exception('IF yes is is a boolean, default can\'t be None.')

  options_items = [f'{"["+k+"]" if default==k else k}={v}'
                                        for k, v in options.items()]
  question = f'{question}\nYour options {",".join(options_items)}:'

  if yes:
    if not quiet:
      # Don't ask, but print to document the default decision.
      print (question, default)
    return default

  while True:
    answer = input(question).strip()
    if answer == '' and default is not None:
      return default
    if answer in options:
      return answer
    # else will ask again


def _format_upstream_yaml (upstream_yaml: YAML, compact: bool = True):
  # removes comments to make it more compact to read
  if compact:
    description = 'upstream configuration (no comments, normalized)'
    content = as_document(upstream_yaml.data, upstream_yaml_schema).as_yaml()
  else:
    description = 'upstream configuration'
    content = upstream_yaml.as_yaml()
  len_top_bars = (58 - len(description)) // 2
  top = f'{"-"*len_top_bars} {description} {"-"*len_top_bars}'
  return (
    f'{top}\n'
    f'{content}'
    f'{"-"*len(top)}'
  )



def _repl_upstream_conf(initial_upstream_conf: str, yes: bool=False
                      , quiet: bool=False, use_template_schema=False):
  if yes:
    raise UserAbortError()
  # repl means "read-eval-print loop"
  editor = _get_editor_command()

  # it would be nice to have a location where the file can be inspected
  # after this program ends, similar to swp files of vim or how git stores
  # such files. However, that should maybe be in the upstream repository
  # rather than in the current working directory. Since I'm undecided
  # I simply go with a temp file
  _tempfilefd, upstream_yaml_file_name = mkstemp(suffix='.yaml'
                                                     , prefix='upstream')
  try:
    # Unlike TemporaryFile(), the user of mkstemp() is responsible for
    # deleting the temporary file when done with it.
    os.close(_tempfilefd)
    print(f'temp file name is {upstream_yaml_file_name}')

    last_good_conf = None
    edit_challenge = initial_upstream_conf

    while True:
      # truncates the file on open
      with open(upstream_yaml_file_name, 'w') as upstream_yaml_file:
        upstream_yaml_file.write(edit_challenge)
      # open it in an editor
      # NOTE the carriage return, this line will be removed again.
      # not sure if this should go to stdout or stderr
      print ('hint: Waiting for your editor to close the file ...'
                                    , end='', flush=True, file=sys.stderr)
      subprocess.run([editor, upstream_yaml_file_name])
      print (CLEARLINE + TOLEFT, end='', flush=True, file=sys.stderr)

      # read the file
      with open(upstream_yaml_file_name, 'r') as upstream_yaml_file:
        updated_upstream_conf = upstream_yaml_file.read()

      # parse the file
      try:
        yaml_schema = upstream_yaml_schema if not use_template_schema \
                                      else upstream_yaml_template_schema
        last_good_conf = dirty_load(updated_upstream_conf, yaml_schema,
                                    allow_flow_style=True)
      except Exception as e:
        answer = user_input(f'The configuration did not parse ({type(e).__name__}):\n\n'
                       f'{e}',
                       OrderedDict(f='fix last edit',
                                   r='retry last edit',
                                   s='start all over',
                                   q='quit program'),
                       # the default should always be an option that does
                       # not require user interaction. That way we can
                       # have an all -y/--no-confirm flag that always
                       # chooses the default.
                       default='q', yes=yes, quiet=quiet)
        if answer == 'f':
          edit_challenge = updated_upstream_conf
        elif answer == 'r':
          # edit_challenge = edit_challenge
          pass
        elif answer == 's':
          edit_challenge = initial_upstream_conf
        else: # anser == 'q':
          raise UserAbortError()
        continue

      return last_good_conf
      # This was thought as an extra check for the user, but I think it's
      # rather anoying than helpful. Note, the user just edited and
      # it parsed successfully.
      # # Ask the user if this looks good.
      # answer = user_input('Use this upstream configuration?\n'
      #       f'{_format_upstream_yaml(last_good_conf)}',
      #       OrderedDict(y='yes',
      #                   e='edit again',
      #                   s='start all over',
      #                   q='quit program'),
      #       default='y', yes=yes, quiet=quiet)
      # if answer == 'y':
      #   return last_good_conf
      # elif answer == 'e':
      #   edit_challenge = last_good_conf.as_yaml()
      # elif answer == 's':
      #   edit_challenge = initial_upstream_conf
      # else: # answer == 'q':
      #   raise UserAbortError()
  finally:
    os.unlink(upstream_yaml_file_name)

def _load_or_repl_upstream(upstream_yaml_text: str, yes: bool = False,
                           quiet: bool = False,
                           use_template_schema: bool = False
                          ) -> typing.Tuple[bool, YAML]:
  try:
    yaml_schema = upstream_yaml_schema if not use_template_schema \
                                    else upstream_yaml_template_schema
    return False, dirty_load(upstream_yaml_text, yaml_schema
                                        , allow_flow_style=True)
  except YAMLValidationError as err:
    answer = user_input('The configuration has schema errors:\n\n'
                     f'{err}',
                     OrderedDict(e='edit',
                                 q='quit program'),
                     default='q', yes=yes, quiet=quiet)
    if answer == 'q':
      raise UserAbortError()
    return True, _repl_upstream_conf(upstream_yaml_text, yes=yes, quiet=quiet,
                                    use_template_schema=use_template_schema)

def _upstream_conf_from_file(filename: str, yes: bool = False,
                                            quiet: bool = False,
                                            use_template_schema: bool = False
                                          ) -> YAML:
  """ If this parses there will be no repl, the user can edit
  the file directly on disk.
  If it doesn't parse, there's a chance to edit until the yaml parses
  and to change the result back to disk.
  """
  with open(filename, 'r+') as upstream_yaml_file:
    upstream_yaml_text = upstream_yaml_file.read()
    edited, upstream_conf_yaml = _load_or_repl_upstream(upstream_yaml_text
                                                  , yes=yes, quiet=quiet
                                                  , use_template_schema=use_template_schema)
    # "edited" is only true when upstream_yaml_text did not parse and
    # was then edited successfully.
    if edited:
      answer = user_input(f'Save changed file {filename}?',
            OrderedDict(y='yes',
                        n='no'),
            default='y', yes=yes, quiet=quiet)
      if answer == 'y':
        upstream_yaml_file.seek(0)
        upstream_yaml_file.truncate()
        upstream_yaml_file.write(upstream_conf_yaml.as_yaml())
  return upstream_conf_yaml


def _upstream_conf_from_scratch(family_name: typing.Union[str, None] = None,
                                yes: bool = False, quiet: bool = False,
                                use_template_schema:bool = False) -> YAML:

  upstream_conf_yaml = dirty_load(upstream_yaml_template, upstream_yaml_template_schema
                                            , allow_flow_style=True)
  if family_name is not None:
    upstream_conf_yaml['name'] = family_name

  if use_template_schema and yes: # for -u/--upstream-yaml
    return upstream_conf_yaml

  template = upstream_conf_yaml.as_yaml()
  return _repl_upstream_conf(template, yes=yes, quiet=quiet,
                                           use_template_schema=use_template_schema)

def _user_input_license(yes: bool=False, quiet: bool=False):
  answer = user_input('To add a new typeface family to Google Fonts we '
              'must know the license of the family.\n'
              'It\'s very likely that OFL is the license that is expected here.',
              OrderedDict(o='OFL: SIL Open Font License',
                          a='Apache License',
                          u='Ubuntu Font License',
                          q='quit program'),
              # the default should always be an option that does
              # not require user interaction. That way we can
              # have an all -y/--no-confirm flag that always chooses the
              # default.
              default='o', yes=yes, quiet=quiet)
  if answer == 'q':
    raise UserAbortError()

  license_dir = {d[0]:d for d in LICENSE_DIRS}[answer]
  return license_dir


def _upstream_conf_from_yaml_metadata(
                              upstream_yaml_text: typing.Union[str, None],
                              metadata_text: typing.Union[str, None],
                              yes: bool = False,
                              quiet: bool = False,
                              use_template_schema: bool = False) -> YAML:
  """ Make a package when the family is in the google/fonts repo.
  Uses data preferred from upstream.yaml, if already present, and fills
  the gaps with data from METADATA.pb. This is to enable the removal of
  redundant data from upstream.yaml when it is in METADATA.pb, while also
  keeping upstream.yaml as the source of truth when in doubt.

  Eventually the common update path.
  """
  upstream_conf = {}
  if metadata_text is not None:
    metadata = fonts_pb2.FamilyProto()
    text_format.Parse(metadata_text, metadata)
    # existing repo, no upstream conf:
    # from METADATA.pb we use:
    #       designer, category, name
    # we won't get **yet**:
    #       source.repository_url
    # we still need the new stuff:
    #       branch, files
    upstream_conf.update({
      'designer': metadata.designer or None,
      'category': list(metadata.category) or None,
      'name': metadata.name  or None,
      # we won't get this just now in most cases!
      'repository_url': metadata.source.repository_url or None,
    })
  if upstream_yaml_text is not None:
    # Only drop into REPL mode if can't parse and validate,
    # and use use_template_schema, because this is not the real deal
    # yet and we can be very forgiving.
    _, upstream_conf_yaml = _load_or_repl_upstream(upstream_yaml_text
                                                  , yes=yes, quiet=quiet
                                                  , use_template_schema=True)

    # remove None values:
    upstream_conf_yaml_data = {k:v for k,v in upstream_conf_yaml.data.items()
                                              if v is not None }
    # Override keys set by METADATA.pb before, if there's overlap.
    upstream_conf.update(upstream_conf_yaml_data)

  upstream_conf_yaml = dirty_load(upstream_yaml_template, upstream_yaml_template_schema
                                       , allow_flow_style=True)
  for k,v in upstream_conf.items():
    if v is None: continue
    upstream_conf_yaml[k] = v

  upstream_yaml_text = upstream_conf_yaml.as_yaml()
  assert upstream_yaml_text is not None

  # two cases:
  # - upstream.yaml may need an update by the user
  # - upstream.yaml may be invalid (updated schema, syntax)
  answer = user_input('Do you want to edit the current upstream configuration?',
                 OrderedDict(y='yes',
                             n='no'),
                 default='n', yes=yes, quiet=quiet)
  if answer == 'y':
    return _repl_upstream_conf(upstream_yaml_text, yes=yes, quiet=quiet ,
                              use_template_schema=use_template_schema)
  _, upstream_conf_yaml =  _load_or_repl_upstream(upstream_yaml_text, yes=yes,
                              quiet=quiet, use_template_schema=use_template_schema)
  return upstream_conf_yaml

def _get_upstream_info(file_or_family: str, is_file: bool, yes: bool,
                          quiet: bool, require_license_dir: bool = True,
                          use_template_schema: bool = False
                          ) -> typing.Tuple[YAML, typing.Union[str, None], dict]:
  # the first task is to acquire an upstream_conf, the license dir and
  # if present the available files for the family in the google/fonts repo.
  license_dir: typing.Union[str, None] = None
  upstream_conf_yaml = None
  gf_dir_content: typing.Dict[str, typing.Dict[str, typing.Any]] = {}

  if not is_file:
    family_name = file_or_family
  else:
    # load a upstream.yaml from disk
    upstream_conf_yaml = _upstream_conf_from_file(file_or_family,
                                      yes=yes, quiet=quiet,
                                      use_template_schema=use_template_schema)
    family_name = upstream_conf_yaml['name'].data

  # TODO:_get_gf_dir_content: is implemented as github graphql query,
  # but, as an alternative, could also be answered with a local
  # clone of the git repository. then _get_gf_dir_content needs a
  # unified api.
  # This could be pereferable if there's a google/fonts clone on the
  # system and e.g. if the user has a bad/no internet access or
  # is running into github api rate limits.
  #
  # if family_name can't be found:
  #    license_dir is None, gf_dir_content is an empty dict
  license_dir, gf_dir_content = _get_gf_dir_content(family_name)

  if license_dir is None:
    # The family is not specified or not found on google/fonts.
    # Can also be an user input error, but we don't handle this yet/here.
    print(f'Font Family "{family_name}" not found on Google Fonts.')
    if require_license_dir:
      license_dir = _user_input_license(yes=yes, quiet=quiet)
    if upstream_conf_yaml is None:
      # if there was no local upstream yaml
      upstream_conf_yaml = _upstream_conf_from_scratch(family_name,
                                        yes=yes, quiet=quiet,
                                        use_template_schema=use_template_schema)
  else:
    print(f'Font Family "{family_name}" is on Google Fonts under "{license_dir}".')

  if upstream_conf_yaml is not None:
    # loaded from_file or created from_scratch
    return upstream_conf_yaml, license_dir, gf_dir_content or {}

  upstream_yaml_text: typing.Union[str, None] = None
  metadata_text: typing.Union[str, None] = None

  if 'upstream.yaml' in gf_dir_content:
    # normal case
    print(f'Using upstream.yaml from google/fonts for {family_name}.')
    file_sha = gf_dir_content['upstream.yaml']['oid']
    response = get_github_gf_blob(file_sha)
    upstream_yaml_text = response.text

  if 'METADATA.pb' in gf_dir_content:
    file_sha = gf_dir_content['METADATA.pb']['oid']
    response = get_github_gf_blob(file_sha)
    metadata_text = response.text

  if upstream_yaml_text is None and metadata_text is None:
    raise Exception('Unexpected: can\'t use google fonts family data '
                    f'for {family_name}.')

  upstream_conf_yaml = _upstream_conf_from_yaml_metadata(upstream_yaml_text,
                                          metadata_text,
                                          yes=yes, quiet=quiet,
                                          use_template_schema=use_template_schema)

  return upstream_conf_yaml, license_dir, gf_dir_content or {}


def _edit_upstream_info(upstream_conf_yaml: YAML, file_or_family: str,
                        is_file: bool, yes:bool, quiet: bool)\
                        -> typing.Tuple[YAML, str, dict]:
  license_dir = None
  gf_dir_content: typing.Dict[str, typing.Dict[str, typing.Any]] = {}

  print(f'Edit upstream conf.')
  upstream_conf_yaml = _repl_upstream_conf(upstream_conf_yaml.as_yaml(),
                                                    yes=yes, quiet=quiet)
  if is_file:
    answer = user_input(f'Save changed file {file_or_family}?',
            OrderedDict(y='yes',
                        n='no'),
            default='y', yes=yes, quiet=quiet)
    if answer == 'y':
      with open(file_or_family, 'w') as upstream_yaml_file:
        upstream_yaml_file.write(upstream_conf_yaml.as_yaml())
  family_name = upstream_conf_yaml['name'].data
  # if family_name can't be found:
  #    license_dir is None, gf_dir_content is an empty dict
  license_dir, gf_dir_content = _get_gf_dir_content(family_name)
  if license_dir is None:
    # The family is not specified or not found on google/fonts.
    # Can also be an user input error, but we don't handle this yet/here.
    print(f'Font Family "{family_name}" not found on Google Fonts.')
    license_dir = _user_input_license(yes=yes, quiet=quiet)

  return upstream_conf_yaml, license_dir, gf_dir_content or {}

def _is_allowed_file(filename: str, no_allowlist: bool=False):
  # there are two places where .ttf files are allowed to go
  # we don't do filename/basename validation here, that's
  # a job for font bakery
  if filename.endswith('.ttf') \
      and os.path.dirname(filename) in ['', 'static']:
    return True # using this!
  elif filename not in ALLOWED_FILES \
                        and not no_allowlist: # this is the default
    return False
  return True

SKIP_NOT_PERMITTED = 'Target is not a permitted filename (see --no_allowlist):'
SKIP_SOURCE_NOT_FOUND = 'Source not found in upstream:'
SKIP_SOURCE_NOT_BLOB = 'Source is not a blob (blob=file):'
SKIP_COPY_EXCEPTION = 'Can\'t copy:'

def _copy_upstream_files_from_git(branch: str, files: dict, repo: pygit2.Repository
            , write_file_to_package: typing.Callable[[str, bytes], None]
            , no_allowlist: bool=False) \
              -> OrderedDict:

  skipped: 'OrderedDict[str, typing.List[str]]' = OrderedDict([
      (SKIP_NOT_PERMITTED, []),
      (SKIP_SOURCE_NOT_FOUND, []),
      (SKIP_SOURCE_NOT_BLOB, []),
      (SKIP_COPY_EXCEPTION, [])
  ])
  for source, target in files.items():
    # else: allow, write_file_to_package will raise errors if target is bad
    if not _is_allowed_file(target, no_allowlist):
      skipped[SKIP_NOT_PERMITTED].append(target)
      continue

    try:
      source_object = repo.revparse_single(f"{branch}:{source}")
    except KeyError as e:
      skipped[SKIP_SOURCE_NOT_FOUND].append(source)
      continue

    if(source_object.type != pygit2.GIT_OBJ_BLOB):
      skipped[SKIP_SOURCE_NOT_BLOB].append(f'{source} (type is {source_object.type_str})')
      continue

    try:
      write_file_to_package(target, source_object.data)
    except Exception as e:
      # i.e. file exists
      skipped[SKIP_COPY_EXCEPTION].append(f'{target} ERROR: {e}')
  # Clean up empty entries in skipped.
  # using list() because we can't delete from a dict during iterated
  for key in list(skipped):
    if not skipped[key]: del skipped[key]
  # If skipped is empty all went smooth.
  return skipped

def _copy_upstream_files_from_dir(source_dir: str, files: dict,
              write_file_to_package: typing.Callable[[str, bytes], None],
              no_allowlist: bool=False) -> OrderedDict:

  skipped: 'OrderedDict[str, typing.List[str]]' = OrderedDict([
      (SKIP_NOT_PERMITTED, []),
      (SKIP_SOURCE_NOT_FOUND, []),
      (SKIP_SOURCE_NOT_BLOB, []),
      (SKIP_COPY_EXCEPTION, [])
  ])
  for source, target in files.items():
    # else: allow, write_file_to_package will raise errors if target is bad
    if not _is_allowed_file(target, no_allowlist):
      skipped[SKIP_NOT_PERMITTED].append(target)
      continue

    try:
      with open(os.path.join(source_dir, source), 'rb') as f:
        source_data = f.read()
    except FileNotFoundError:
      skipped[SKIP_SOURCE_NOT_FOUND].append(source)
      continue
    except Exception as err:
      # e.g. IsADirectoryError
      skipped[SKIP_SOURCE_NOT_BLOB].append(f'{source} ({type(err).__name__})')
      continue

    try:
      write_file_to_package(target, source_data)
    except Exception as e:
      # i.e. file exists
      skipped[SKIP_COPY_EXCEPTION].append(f'{target} ERROR: {e}')

  # Clean up empty entries in skipped.
  # using list() because we can't delete from a dict during iterated
  for key in list(skipped):
    if not skipped[key]: del skipped[key]
  # If skipped is empty all went smooth.
  return skipped

def _create_or_update_metadata_pb(upstream_conf: YAML,
                                  tmp_package_family_dir:str,
                                  upstream_commit_sha:str) -> None:
  metadata_file_name = os.path.join(tmp_package_family_dir, 'METADATA.pb')
  try:
    subprocess.run(['gftools', 'add-font', tmp_package_family_dir]
                                , check=True, stdout=subprocess.PIPE
                                , stderr=subprocess.PIPE)
  except subprocess.CalledProcessError as e:
    print(str(e.stderr, 'utf-8'), file=sys.stderr)
    raise e

  metadata = fonts_pb2.FamilyProto()

  with open(metadata_file_name, 'rb') as fb:
    text_format.Parse(fb.read(), metadata)

  # make upstream_conf the source of truth for some entries
  metadata.name = upstream_conf['name']
  for font in metadata.fonts:
    font.name = upstream_conf['name']
  metadata.designer = upstream_conf['designer']

  metadata.category[:] = upstream_conf['category']

  # metadata.date_added # is handled well

  metadata.source.repository_url = upstream_conf['repository_url']
  metadata.source.commit = upstream_commit_sha

  language_comments = fonts.LanguageComments(LoadLanguages())
  fonts.WriteProto(metadata, metadata_file_name, comments=language_comments)

def _create_package_content(package_target_dir: str, repos_dir: str,
        upstream_conf_yaml: YAML, license_dir: str, gf_dir_content:dict,
        allow_build: bool, yes: bool, quiet: bool,
        no_allowlist: bool = False) -> str:
  print(f'Creating package with \n{_format_upstream_yaml(upstream_conf_yaml)}')
  upstream_conf = upstream_conf_yaml.data
  upstream_commit_sha = None

  family_name_normal = _family_name_normal(upstream_conf['name'])
  family_dir = os.path.join(license_dir, family_name_normal)
  package_family_dir = os.path.join(package_target_dir, family_dir)
  # putting state into functions, could be done with classes/methods as well
  write_file_to_package = functools.partial(_write_file_to_package,
                                            package_family_dir)
  file_in_package = functools.partial(_file_in_package,
                                      package_family_dir)
  # Get and add upstream files!
  upstream_dir_target = (
      f'{upstream_conf["repository_url"]}'
      f'__{upstream_conf["branch"]}'
      # Despite of '.' and '/' I'd expect the other replacements
      # not so important in this case.
    ) \
    .replace('://', '_') \
    .replace('/', '_') \
    .replace('.', '_') \
    .replace('\\', '_')

  local_repo_path_marker = 'local://'
  if upstream_conf['repository_url'].startswith(local_repo_path_marker):
    print(f'WARNING using "local://" hack for repository_url: {upstream_conf["repository_url"]}')
    local_path = upstream_conf['repository_url'][len(local_repo_path_marker):]
    upstream_dir = os.path.expanduser(local_path)
  else:
    upstream_dir = os.path.join(repos_dir, upstream_dir_target)
    if not os.path.exists(upstream_dir):
      # for super families it's likely that we can reuse the same clone
      # of the repository for all members
      _shallow_clone_git(upstream_dir, upstream_conf['repository_url']
                                    , upstream_conf['branch'])
  repo = pygit2.Repository(upstream_dir)

  upstream_commit = repo.revparse_single(upstream_conf['branch'])
  upstream_commit_sha = upstream_commit.hex

  # Copy all files from upstream_conf['files'] to package_family_dir
  # We are strict about what to allow, unexpected files
  # are not copied. Instead print a warning an suggest filing an
  # issue if the file is legitimate. A flag to explicitly
  # skip the allowlist check (--no_allowlist)
  # enables making packages even when new, yet unknown files are required).
  # Do we have a Font Bakery check for expected/allowed files? Would
  # be a good complement.
  if upstream_conf['build']:

    print(f'Found build command:\n  $ {upstream_conf["build"]}')
    if not allow_build:
      answer = user_input(f'Can\'t execute build command without explicit '
              'permission. Don\'t allow this lightly '
              'and review build command, build process and its dependencies prior. '
              'This support for building from sources is provisional, a '
              'discussion can be found at https://github.com/googlefonts/gftools/issues/231',
              OrderedDict(b='build',
                          q='quit program'),
              default='q', yes=yes, quiet=quiet)
      if answer == 'q':
        raise UserAbortError('Can\'t execute required build command. '
                              'Use --allow-build to allow explicitly.')
    with TemporaryDirectory() as tmp:
      print(f'Building...')
      subprocess.run(['git', 'clone', upstream_dir, tmp], check=True)
      subprocess.run(['bash', '-c', upstream_conf['build']]
                       , cwd=tmp
                       , check=True)
      print(f'DONE building!')
      skipped = _copy_upstream_files_from_dir(tmp, upstream_conf['files'],
                        write_file_to_package, no_allowlist=no_allowlist)
  else:
    skipped = _copy_upstream_files_from_git(upstream_conf['branch'],
                    upstream_conf['files'], repo, write_file_to_package,
                    no_allowlist=no_allowlist)
  if skipped:
    message = ['Some files from upstream_conf could not be copied.']
    for reason, items in skipped.items():
      message.append(reason)
      for item in items:
        message.append(f' - {item}')
    # The allowlist can be ignored using the flag no_allowlist flag,
    # but the rest should be fixed in the files map, because it's
    # obviously wrong, not working, configuration.
    # TODO: This case could (but should it?) be a repl-case to ask
    # interactively, if the no_allowlist flag should be used then,
    # if yes, _copy_upstream_files could be tried again. But given
    # that the use case for the flag is a narrow one, I doubt the
    # effort needed and the added complexity is worth it.
    raise ProgramAbortError('\n'.join(message))

  # Get and add all files from google/fonts
  for name, entry in gf_dir_content.items():
    # not copying old TTFs, directories and files that are already there
    if name.endswith('.ttf') \
          or entry['type'] != 'blob'\
          or file_in_package(name):
      continue
    file_sha = gf_dir_content[name]['oid']
    response = get_github_gf_blob(file_sha)
    write_file_to_package(name, response.content)

  # create/update METADATA.pb
  _create_or_update_metadata_pb(upstream_conf, package_family_dir,
                                upstream_commit_sha)

  # create/update upstream.yaml
  # Remove keys that are also in METADATA.pb googlefonts/gftools#233
  # and also clear all comments.
  redundant_keys = {'name', 'category', 'designer', 'repository_url'}
  upstream_conf_stripped = OrderedDict((k, v) for k, v in upstream_conf.items() \
                                                  if k not in redundant_keys)
  # Don't keep an empty build key.
  if 'build' in upstream_conf_stripped and (upstream_conf_stripped['build'] == ''\
                or upstream_conf_stripped['build'] is None):
    del upstream_conf_stripped['build']
  upstream_conf_stripped_yaml = as_document(upstream_conf_stripped, upstream_yaml_stripped_schema)
  with open(os.path.join(package_family_dir, 'upstream.yaml'), 'w') as f:
    f.write(upstream_conf_stripped_yaml.as_yaml())
  print(f'DONE Creating package for {upstream_conf["name"]}!')
  return family_dir

def _check_git_target(target: str) -> None:
  try:
    repo = repo = pygit2.Repository(target)
  except Exception as e:
    raise ProgramAbortError(f'Can\'t open "{target}" as git repository. '
                            f'{e} ({type(e).__name__}).')
  repo_owner = 'google'
  repo_name = 'fonts'
  repo_name_with_owner = f'{repo_owner}/{repo_name}'
  remote_name_or_none = _find_github_remote(repo, repo_owner, repo_name, 'main')
  if remote_name_or_none is None:
    # NOTE: we could ask the user if we should add the missing remote.
    # This makes especially sense if the repository is a fork of
    # google/fonts and essentially has the same history/object database.
    # It would be very uncommon, probably unintended, if the repo is not
    # related to the google/fonts repo and fetching from that remote would
    # have to download a lot of new data, as well as probably create
    # confusing situations for the user when dealing with GitHub PRs etc.
    print (f'The git repository at target "{target}" has no remote for '
      f'GitHub {repo_name_with_owner}.\n'
      'You can add it by running:\n'
      f'$ cd {target}\n'
      f'$ git remote add googlefonts {GITHUB_REPO_SSH_URL(repo_name_with_owner=repo_name_with_owner)}.git\n'
      'For more context, run:\n'
      '$ git remote help')

    raise ProgramAbortError(f'The target git repository has no remote for '
              f'GitHub google/fonts.')

def _check_directory_target(target: str) -> None:
  if not os.path.isdir(target):
    raise ProgramAbortError(f'Target "{target}" is not a directory.')

def _check_target(is_gf_git: bool, target: str) -> None:
  if is_gf_git:
    return _check_git_target(target)
  else:
    return _check_directory_target(target)

def _git_tree_from_dir(repo: pygit2.Repository, tmp_package_family_dir: str) -> str:
  trees: typing.Dict[str, str] = {}
  for root, dirs, files in os.walk(tmp_package_family_dir, topdown=False):
    # if root == tmp_package_family_dir: rel_dir = '.'
    rel_dir = os.path.relpath(root, tmp_package_family_dir)
    treebuilder = repo.TreeBuilder()
    for filename in files:
      with open(os.path.join(root, filename), 'rb') as f:
        blob_id = repo.create_blob(f.read())
      treebuilder.insert(filename, blob_id, pygit2.GIT_FILEMODE_BLOB)
    for dirname in dirs:
      path = dirname if rel_dir == '.' else os.path.join(rel_dir, dirname)
      tree_id = trees[path]
      treebuilder.insert(dirname, tree_id, pygit2.GIT_FILEMODE_TREE)
    # store for use in later iteration, note, we're going bottom up
    trees[rel_dir] = treebuilder.write()
  return trees['.']

def _git_write_file(repo: pygit2.Repository, tree_builder: pygit2.TreeBuilder,
                                        file_path: str, data: str) -> None:
  blob_id = repo.create_blob(data)
  return _git_makedirs_write(repo, tree_builder, PurePath(file_path).parts,
                            blob_id, pygit2.GIT_FILEMODE_BLOB)


def _git_makedirs_write(repo: pygit2.Repository, tree_builder:pygit2.TreeBuilder,
          path:typing.Tuple[str, ...], git_obj_id: str, git_obj_filemode:int) -> None:
  name, rest_path = path[0], path[1:]
  if not rest_path:
    tree_builder.insert(name, git_obj_id, git_obj_filemode)
    return

  child_tree = tree_builder.get(name)
  try:
    child_tree_builder = repo.TreeBuilder(child_tree)
  except TypeError as e:
    # will raise TypeError if license_dir_tree is None i.e. not exisiting
    # but also if child_tree is not a pygit2.GIT_FILEMODE_TREE

    # os.makedirs(name, exists_ok=True) would raise FileExistsError if
    # it is tasked to create a directory where a file already exists
    # It seems unlikely that we want to override existing files here
    # so I copy that behavior.
    if child_tree is not None:
      # FileExistsError is an OSError so it's probably misused here
      raise FileExistsError(f'The git entry {name} exists as f{child_tree.type_str}.')
    child_tree_builder = repo.TreeBuilder()

  _git_makedirs_write(repo, child_tree_builder, rest_path, git_obj_id, git_obj_filemode)
  child_tree_id = child_tree_builder.write()
  tree_builder.insert(name, child_tree_id,  pygit2.GIT_FILEMODE_TREE)

def _git_copy_dir(repo: pygit2.Repository, tree_builder: pygit2.TreeBuilder,
                      source_dir:str, target_dir:str) -> None:
  # This is a new tree, i.e. not based on an existing tree.
  tree_id = _git_tree_from_dir(repo, source_dir)

  # Here we insert into an existing tree.
  _git_makedirs_write(repo, tree_builder, PurePath(target_dir).parts,
                            tree_id, pygit2.GIT_FILEMODE_TREE)

# thanks https://stackoverflow.com/a/1094933
def _sizeof_fmt(num, suffix='B'):
  for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
    if abs(num) < 1024.0:
      return "%3.1f%s%s" % (num, unit, suffix)
    num /= 1024.0
  return "%.1f%s%s" % (num, 'Yi', suffix)

def _push(repo: pygit2.Repository, url: str, local_branch_name: str,
          remote_branch_name: str, force: bool):
  full_local_ref = local_branch_name if local_branch_name.find('refs/') == 0 \
                                     else f'refs/heads/{local_branch_name}'
  full_remote_ref = f'refs/heads/{remote_branch_name}'
  ref_spec = f'{full_local_ref}:{full_remote_ref}'
  if force:
    # ref_spec for force pushing must include a + at the start.
    ref_spec = f'+{ref_spec}'

  # NOTE: pushing using pygit2 is currently not working on MacOS, this is
  # related to SSH issues. Here's a traceback:
  #                   https://github.com/googlefonts/gftools/issues/238
  # Since we did it already once with `git clone --depth 1`, this is also
  # being worked around by using the CLI git directly.
  #
  # callbacks = PYGit2RemoteCallbacks()
  # with _create_tmp_remote(repo, url) as remote:
  #   # https://www.pygit2.org/remotes.html#pygit2.Remote.push
  #   # When the remote has a githook installed, that denies the reference
  #   # this function will return successfully. Thus it is strongly recommended
  #   # to install a callback, that implements RemoteCallbacks.push_update_reference()
  #   # and check the passed parameters for successfull operations.
  #
  #
  #   remote.push([ref_spec], callbacks=callbacks)
  subprocess.run(['git', 'push', url, ref_spec],
                    cwd=repo.path,
                    check=True,
                    stdout=subprocess.PIPE)

  #if callbacks.rejected_push_message is not None:
  #  raise Exception(callbacks.rejected_push_message)

GITHUB_REPO_SSH_URL = 'git@github.com:{repo_name_with_owner}.git'.format


def _make_pr(repo: pygit2.Repository, local_branch_name: str,
                                  pr_upstream: str, push_upstream: str,
                                  pr_title: str, pr_message_body: str):
  print('Making a Pull Request …')
  if not push_upstream:
    push_upstream = pr_upstream

  push_owner, _push_repo = push_upstream.split('/')
  pr_owner, pr_repo = pr_upstream.split('/')
  url = GITHUB_REPO_SSH_URL(repo_name_with_owner=push_upstream)

  remote_branch_name = local_branch_name
  # We must only allow force pushing/general pushing to branch names that
  # this tool *likely* created! Otherwise, we may end up force pushing
  # to `main` branch! Hence: we expect a prefix for remote_branch_name indicating
  # this tool created it.
  if remote_branch_name.find(GIT_NEW_BRANCH_PREFIX) != 0:
    remote_branch_name = (f'{GIT_NEW_BRANCH_PREFIX}'
                          f'{remote_branch_name.replace(os.sep, "_")}')

  print('git push:\n'
          f'  url is {url}\n'
          f'  local branch name is {local_branch_name}\n'
          f'  remote branch name is {remote_branch_name}\n'
  )
  # Always force push?
  # If force == False and we update an existing remote:
  #   _pygit2.GitError: cannot push non-fastforwardable reference
  # But I don't use the --force flag here, because I believe this is
  # very much the standard case, i.e. that we update existing PRs.
  _push(repo, url, local_branch_name, remote_branch_name, force=True)
  print('DONE git push!')

  pr_head = f'{push_owner}:{remote_branch_name}'
  pr_base_branch = 'main'  # currently we always do PRs to main
    #_updateUpstream(prRemoteName, prRemoteRef))
    #// NOTE: at this point the PUSH was already successful, so the branch
    #// of the PR exists or if it existed it has changed.
  client = GitHubClient(pr_owner, pr_repo)
  open_prs = client.open_prs(pr_head, pr_base_branch)

  if not len(open_prs):
    # No open PRs, creating …
    result = client.create_pr(pr_title, pr_message_body, pr_head, pr_base_branch)
    print(f'Created a PR #{result["number"]} {result["html_url"]}')
  else:
    # found open PR
    pr_issue_number = open_prs[0]['number']
    pr_comment_body = f'Updated\n\n## {pr_title}\n\n---\n{pr_message_body}'
    result = client.create_issue_comment(pr_issue_number, pr_comment_body)
    print(f'Created a comment in PR #{pr_issue_number} {result["html_url"]}')

def _get_root_commit(repo: pygit2.Repository, base_remote_branch:str,
                                  tip_commit: pygit2.Commit) -> pygit2.Commit:
  for root_commit in repo.walk(tip_commit.id,
                  pygit2.GIT_SORT_TOPOLOGICAL | pygit2.GIT_SORT_TIME):
    try:
      # If it doesn't raise KeyError i.e. root_commit is contained in
      # our base_remote_branch and not part of the PR.
      return repo.branches.remote.with_commit(root_commit)[base_remote_branch].peel()
    except KeyError:
      continue
    break;

def _get_change_info_from_diff(repo: pygit2.Repository, root_tree: pygit2.Tree,
                                            tip_tree: pygit2.Tree) -> typing.Dict:
  # I probably also want the changed files between root_commit and tip commit
  diff =  repo.diff(root_tree, tip_tree)
  all_touched_files = set()
  for delta in diff.deltas:
    # possible status chars:
    #   GIT_DELTA_ADDED:      A
    #   GIT_DELTA_DELETED:    D
    #   GIT_DELTA_MODIFIED:   M
    #   GIT_DELTA_RENAMED:    R
    #   GIT_DELTA_COPIED:     C
    #   GIT_DELTA_IGNORED:    I
    #   GIT_DELTA_UNTRACKED:  ?
    #   GIT_DELTA_TYPECHANGE: T
    #   GIT_DELTA_UNREADABLE: X
    #   default:              ' '
    if delta.status_char() == 'D':
      all_touched_files.add(delta.old_file.path)
    else:
      all_touched_files.add(delta.new_file.path)
  touched_family_dirs = set()
  for filename in all_touched_files:
    for dirname in LICENSE_DIRS:
      if filename.startswith(f'{dirname}{os.path.sep}'):
        # items are e.g. ('ofl', 'gelasio')
        touched_family_dirs.add(tuple(filename.split(os.path.sep)[:2]))
        break
  family_changes_dict = {}
  for dir_path_tuple in touched_family_dirs:
    family_tree: pygit2.Tree = tip_tree
    for pathpart in dir_path_tuple:
      family_tree = family_tree / pathpart

    metadata_blob: pygit2.Blob = family_tree / 'METADATA.pb'
    metadata = fonts_pb2.FamilyProto()
    text_format.Parse(metadata_blob.data, metadata)

    # get the version
    first_font_file_name = metadata.fonts[0].filename
    first_font_blob: pygit2.Blob = family_tree / first_font_file_name
    first_font_file = BytesIO(first_font_blob.data)
    ttFont = TTFont(first_font_file)
    version:typing.Union[None, str] = None
    NAME_ID_VERSION = 5
    for entry in ttFont['name'].names:
      if entry.nameID == NAME_ID_VERSION:
        # just taking the first instance
        version = entry.string.decode(entry.getEncoding())
        if version:
          break

    # repoNameWithOwner
    prefix  = 'https://github.com/'
    suffix = '.git'
    repoNameWithOwner: typing.Union[None,str] = None
    if metadata.source.repository_url.startswith(prefix):
      repoNameWithOwner = '/'.join(metadata.source.repository_url[len(prefix):]
                                           .split('/')[0:2])
      if repoNameWithOwner.endswith(suffix):
        repoNameWithOwner = repoNameWithOwner[:-len(suffix)]
    commit_url: typing.Union[None,str] = None
    if repoNameWithOwner:
      commit_url = f'https://github.com/{repoNameWithOwner}/commit/{metadata.source.commit}'

    family_changes_dict['/'.join(dir_path_tuple)] = {
      'family_name': metadata.name,
      'repository': metadata.source.repository_url,
      'commit': metadata.source.commit,
      'version': version or '(unknown version)',
      'repoNameWithOwner': repoNameWithOwner,
      'commit_url': commit_url
    }
  return family_changes_dict

def _title_message_from_diff(repo: pygit2.Repository, root_tree: pygit2.Tree,
                            tip_tree: pygit2.Tree) -> typing.Tuple[str, str]:
  family_changes_dict = _get_change_info_from_diff(repo, root_tree, tip_tree)
  title = []
  body = []
  for _, fam_entry in family_changes_dict.items():
    title.append(f'{fam_entry["family_name"]}: {fam_entry["version"]} added')
    commit = fam_entry['commit_url'] or fam_entry['commit']
    body.append(f'* {fam_entry["family_name"]} '
               f'{fam_entry["version"]} taken from the upstream repo '
               f'{fam_entry["repository"]} at commit {commit}.'
               )
  return '; '.join(title), '\n'.join(body)

def _git_get_path(tree: pygit2.Tree, path: str) -> pygit2.Object:
  last = tree
  for pathpart in PurePath(path).parts:
    last = last / pathpart
  return last

def _git_make_commit(repo: pygit2.Repository, add_commit: bool, force: bool,
          yes: bool, quiet: bool, local_branch: str, remote_name: str,
          base_remote_branch: str, tmp_package_family_dir: str,
          family_dir: str):
  base_commit = None
  if add_commit:
    try:
      base_commit = repo.branches.local[local_branch].peel()
    except KeyError as e:
      pass

  if not base_commit:
    #fetch! make sure we're on the actual gf main HEAD
    _git_fetch_main(repo, remote_name)
    # base_commit = repo.revparse_single(f'refs/remotes/{base_remote_branch}')
    # same but maybe better readable:
    base_commit = repo.branches.remote[base_remote_branch].peel()

  # Maybe I can start with the commit tree here ...
  treeBuilder = repo.TreeBuilder(base_commit.tree)
  _git_copy_dir(repo, treeBuilder, tmp_package_family_dir, family_dir)

  # create the commit
  user_name = list(repo.config.get_multivar('user.name'))[0]
  user_email = list(repo.config.get_multivar('user.email'))[0]
  author = pygit2.Signature(user_name, user_email)
  committer = pygit2.Signature(user_name, user_email)

  new_tree_id = treeBuilder.write()
  new_tree: pygit2.Tree = repo.get(new_tree_id)
  title, body = _title_message_from_diff(repo, base_commit.tree, new_tree)

  commit_id = repo.create_commit(
          None,
          author, committer,
          f'[gftools-packager] {title}\n\n{body}',
          new_tree_id,
          [base_commit.id] # parents
  )

  commit = repo.get(commit_id)
  # create branch or add to an existing one if add_commit
  while True:
    try:
      repo.branches.local.create(local_branch, commit, force=add_commit or force)
    except pygit2.AlreadyExistsError:
      # _pygit2.AlreadyExistsError: failed to write reference
      #     'refs/heads/gftools_packager_ofl_gelasio': a reference with
      #     that name already exists.
      answer = user_input(f'Can\'t override existing branch {local_branch}'
                          ' without explicit permission.',
              OrderedDict(f='force override',
                          q='quit program'),
              default='q', yes=yes, quiet=quiet)
      if answer == 'q':
        raise UserAbortError(f'Can\'t override existing branch {local_branch}. '
                            'Use --branch to specify another branch name. '
                            'Use --force to allow explicitly.')
      else: # answer == 'f'
        force = True
        continue
    break

  # only for reporting
  target_label = f'git branch {local_branch}'
  package_contents = []
  for root, dirs, files in _git_tree_walk(family_dir, commit.tree):
    for filename in files:
      entry_name = os.path.join(root, filename)
      filesize = commit.tree[entry_name].size
      package_contents.append((entry_name, filesize))
  _print_package_report(target_label, package_contents)

def _packagage_to_git(tmp_package_family_dir: str, target: str, family_dir: str,
                     branch: str, force:bool, yes: bool, quiet: bool,
                     add_commit: bool) -> None:

  repo = pygit2.Repository(target)
  # we checked that it exists earlier!
  remote_name = _find_github_remote(repo, 'google', 'fonts', 'main')
  base_remote_branch = f'{remote_name}/main'
  if remote_name is None:
    raise Exception('No remote found for google/fonts main.')

  _git_make_commit(repo, add_commit, force, yes, quiet, branch,
                   remote_name, base_remote_branch, tmp_package_family_dir,
                   family_dir)


def _dispatch_git(target: str, target_branch: str,pr_upstream: str,
                  push_upstream: str) -> None:
  repo = pygit2.Repository(target)
  # we checked that it exists earlier!
  remote_name = _find_github_remote(repo, 'google', 'fonts', 'main')
  base_remote_branch = f'{remote_name}/main'
  if remote_name is None:
    raise Exception('No remote found for google/fonts main.')

  git_branch: pygit2.Branch = repo.branches.local[target_branch]
  tip_commit: pygit2.Commit = git_branch.peel()
  root_commit: pygit2.Commit = _get_root_commit(repo, base_remote_branch, tip_commit)
  pr_title, _ = _title_message_from_diff(repo, root_commit.tree, tip_commit.tree)
  if not pr_title:
    # Happens e.g. if we have a bunch of commits that revert themselves,
    # to me this happened in development, in a for production use very unlikely
    # situation.
    # But can also happen if we PR commits that don't do changes in family
    # dirs. In these cases the PR author should probably come up with a
    # better title than this placeholder an change it in the GitHub web-GUI.
    pr_title = '(UNKNOWN gftools-packager: found no family changes)'

  current_commit = tip_commit
  messages = []
  while current_commit.id != root_commit.id:
    messages.append(f' {current_commit.short_id}: {current_commit.message}')
    current_commit = current_commit.parents[0]
  pr_message_body  = '\n\n'.join(reversed(messages))

  _make_pr(repo, target_branch, pr_upstream, push_upstream,
                                            pr_title, pr_message_body)

def _packagage_to_dir(tmp_package_family_dir: str, target: str,
                  family_dir: str, force: bool, yes: bool, quiet: bool):
  # target is a directory:
  target_family_dir = os.path.join(target, family_dir)
  if os.path.exists(target_family_dir):
    if not force:
      answer = user_input(f'Can\'t override existing directory {target_family_dir}'
                          ' without explicit permission.',
              OrderedDict(f='force override',
                          q='quit program'),
              default='q', yes=yes, quiet=quiet)
      if answer == 'q':
        raise UserAbortError('Can\'t override existing directory '
                              f'{target_family_dir}. '
                              'Use --force to allow explicitly.')
    shutil.rmtree(target_family_dir)
  else: # not exists
    os.makedirs(os.path.dirname(target_family_dir), exist_ok=True)
  shutil.move(tmp_package_family_dir, target_family_dir)

  # only for reporting
  target_label = f'directory {target}'
  package_contents = []
  for root, dirs, files in os.walk(target_family_dir):
    for filename in files:
      full_path = os.path.join(root, filename)
      entry_name = os.path.relpath(full_path, target)
      filesize = os.path.getsize(full_path)
      package_contents.append((entry_name, filesize))
  print(f'Package Directory: {target_family_dir}')
  _print_package_report(target_label, package_contents)

def _write_upstream_yaml_backup(upstream_conf_yaml: YAML) -> str:
  family_name_normal = _family_name_normal(upstream_conf_yaml['name'].data)
  count = 0
  while True:
    counter = '' if count == 0 else f'_{count}'
    filename = f'./{family_name_normal}.upstream{counter}.yaml'
    try:
      # 'x': don't override existing files
      with open(filename, 'x') as f:
        f.write(upstream_conf_yaml.as_yaml())
    except FileExistsError:
      # retry until the file could be created, file name changes
      count += 1
      continue
    break
  return filename

def _packages_to_target(tmp_package_dir: str, family_dirs: typing.List[str],
                        target: str, is_gf_git: bool,
                        branch: str, force: bool,
                        yes: bool, quiet: bool, add_commit: bool) ->None:
  for i, family_dir in enumerate(family_dirs):
    tmp_package_family_dir = os.path.join(tmp_package_dir, family_dir)
    # NOTE: if there are any files outside of family_dir that need moving
    # that is not yet supported! The reason is, there's no case for this
    # yet. So, if _create_package_content is changed to put files outside
    # of family_dir, these targets will have to follow and implement it.
    if is_gf_git:
      if i > 0:
        add_commit = True
      _packagage_to_git(tmp_package_family_dir, target, family_dir,
                               branch, force, yes, quiet, add_commit)
    else:
      _packagage_to_dir(tmp_package_family_dir, target, family_dir,
                               force, yes, quiet)


def _branch_name_from_family_dirs(family_dirs: typing.List[str]) -> str:
  if len(family_dirs) == 1:
    return f'{GIT_NEW_BRANCH_PREFIX}{family_dirs[0].replace(os.sep, "_")}'

  by_licensedir: typing.Dict[str, typing.List[str]] = {};
  for f in family_dirs:
    license_dir = os.path.dirname(f)
    if license_dir not in by_licensedir:
      by_licensedir[license_dir] = []
    by_licensedir[license_dir].append(os.path.basename(f))

  # All the sorting is to achieve the same branch name, when
  # family_dirs comes in a different order but with the same content.
  particles = []
  for license_dir, families in by_licensedir.items():
    particles.append(f'{license_dir}_{"-".join(sorted(families))}')

  # Could be like (in an extreme case):
  # gftools_packager_apache_arimo-cherrycreamsoda_ofl_acme-balsamiqsans-librebarcode39extendedtext
  full_branch_name = f'{GIT_NEW_BRANCH_PREFIX}{"_".join(sorted(particles))}'
  # I don't know hard limits here
  max_len = 60
  if len(full_branch_name) <= max_len:
    return full_branch_name
  hash_hex_ini = sha1(full_branch_name.encode('utf-8')).hexdigest()[:10]
  # This is the shortened version from above:
  # gftools_packager_apache_arimo-cherrycreamsoda_ofl_d79615d347
  return f'{full_branch_name[:max_len-11]}_{hash_hex_ini}'

def _file_or_family_is_file(file_or_family: str) -> bool:
  return file_or_family.endswith('.yaml') or \
         file_or_family.endswith('.yml') # .yml is common, too

def _output_upstream_yaml(file_or_family: typing.Union[str, None], target: str,
                    yes: bool, quiet: bool, force: bool) -> None:
  if not file_or_family:
     # just use the template
     upstream_conf_yaml =  dirty_load(upstream_yaml_template, upstream_yaml_template_schema
                                            , allow_flow_style=True)
  else:
    is_file = _file_or_family_is_file(file_or_family)
    upstream_conf_yaml, _, _ = _get_upstream_info(file_or_family, is_file,
                                    yes, quiet, require_license_dir=False,
                                    use_template_schema=True)
  # save!
  while True:
    try:
      with open(target, 'x' if not force else 'w') as f:
        f.write(upstream_conf_yaml.as_yaml())
      break
    except FileExistsError:
      if not force:
        answer = user_input(f'Can\'t override existing target file {target}'
                          ' without explicit permission.',
              OrderedDict(f='force override',
                          q='quit program'),
              default='q', yes=yes, quiet=quiet)
      if answer == 'q':
        raise UserAbortError('Can\'t override existing target file '
                              f'{target}. '
                              'Use --force to allow explicitly.')
      else: # answer == 'f'
        force = True
        continue
  print(f'DONE upstream conf saved as {target}!')


def make_package(file_or_families: typing.List[str], target: str, yes: bool,
                 quiet: bool, no_allowlist: bool, is_gf_git: bool, force: bool,
                 add_commit: bool, pr: bool, pr_upstream: str,
                 push_upstream: str, upstream_yaml: bool,
                 allow_build: bool, branch: typing.Union[str, None]=None):

  if upstream_yaml:
    return _output_upstream_yaml(file_or_families[0] if file_or_families else None,
                            target, yes, quiet, force)
  # some flags can be set implicitly
  pr = pr or bool(push_upstream) or bool(pr_upstream)
  # set default
  if not pr_upstream: pr_upstream = 'google/fonts'

  is_gf_git = is_gf_git or bool(branch) or add_commit or pr
  # Basic early checks. Raises if target does not qualify.
  _check_target(is_gf_git, target)

  # note: use branch if it is explicit (and if is_gf_git)
  target_branch = branch if branch is not None else ''

  family_dirs: typing.List[str] = []
  with TemporaryDirectory() as tmp_dir:
    tmp_package_dir = os.path.join(tmp_dir, 'packages')
    os.makedirs(tmp_package_dir, exist_ok=True)
    tmp_repos_dir = os.path.join(tmp_dir, 'repos')
    os.makedirs(tmp_repos_dir, exist_ok=True)

    for file_or_family in file_or_families:
      is_file = _file_or_family_is_file(file_or_family)
      edit = False
      while True: # repl
        if not edit:
          ( upstream_conf_yaml, license_dir,
            gf_dir_content ) = _get_upstream_info(file_or_family, is_file,
                                                                  yes, quiet)
        else:
          ( upstream_conf_yaml, license_dir,
            gf_dir_content ) = _edit_upstream_info(upstream_conf_yaml,
                                        file_or_family, is_file, yes, quiet)
          edit = False # reset
        assert isinstance(license_dir, str)
        try:
          family_dir = _create_package_content(tmp_package_dir, tmp_repos_dir,
                                upstream_conf_yaml, license_dir,
                                gf_dir_content, allow_build, yes, quiet,
                                no_allowlist)
          family_dirs.append(family_dir)
        except UserAbortError as e:
          # The user aborted already, no need to bother any further.
          # FIXME: however, we don't get to the point where we can save
          # the upstream conf to disk, and that may be desirable here!
          raise e
        except Exception:
          error_io = StringIO()
          traceback.print_exc(file=error_io)
          error_io.seek(0)
          answer = user_input(f'Upstream conf caused an error:'
                              f'\n-----\n\n{error_io.read()}\n-----\n'
                              'How do you want to proceed?',
                  OrderedDict(e='edit upstream conf and retry',
                              q='raise and quit program'),
                  default='q', yes=yes, quiet=quiet)
          if answer == 'q':
            if not yes:
              # Should be possible to save to original file if is_file
              # but we should give that option only if the file would change.
              # Also, in edit_upstream_info it is possible to save to the
              # original file.
              answer = user_input('Save upstream conf to disk?\nIt can be '
                                   'annoying having to redo all changes, which '
                                   'will be lost if you choose no.\n'
                                   'The saved file can be edited and used with '
                                   'the --file option.' ,
                  OrderedDict(y='yes—save to disk',
                              n='no—discard changes'),
                  default='y', yes=yes, quiet=quiet)
              if answer == 'y':
                upstream_yaml_backup_filename = _write_upstream_yaml_backup(
                                                          upstream_conf_yaml)
                print(f'Upstream conf has been saved to: {upstream_yaml_backup_filename}')
            raise UserAbortError()
          else:
            # answer == 'e'
            # continue loop: go back to _get_upstream_info
            edit = True
            continue
        # Done with file_or_family!
        break # break the REPL while loop.
    if not family_dirs:
      print('No families to package.')
    # done with collecting data for all file_or_families

    if is_gf_git:
      # need to have a unified branch for all file_or_families ...
      # if there are more than one families ...
      if not branch:
        target_branch = _branch_name_from_family_dirs(family_dirs)
    _packages_to_target(tmp_package_dir, family_dirs, target, is_gf_git,
                        target_branch, force, yes, quiet, add_commit)

  if pr and is_gf_git:
    _dispatch_git(target, target_branch, pr_upstream, push_upstream)


def _print_package_report (target_label: str,
            package_contents: typing.List[typing.Tuple[str, int]]) -> None:
  print(f'Created files in {target_label}:')
  for entry_name, filesize in package_contents:
    filesize_str = filesize
    print(f'   {entry_name} {_sizeof_fmt(filesize_str)}')

def _find_github_remote(repo: pygit2.Repository, owner: str, name: str,
          branch: typing.Union[str, None] = None) -> typing.Union[str, None]:
  """
  Find a remote-name that is a good fit for the GitHub owner and repo-name.
  A good fit is when we can use it to fetch/push from/to GitHubs repository
  to/from the `branch` if branch is given or any branch if `branch` is None.

  Returns remote-name or None
  """
  searched_repo = f'{owner}/{name}'
  # If we plan to also push to these, it is important which
  # remote url we choose, esp. because of authentication methods.
  # I'd try to pick remotes after the order below, e.g. first the
  # ssh based urls, then the https url, because human users will
  # likely have a working ssh setup if they are pushing to github,
  # An environment like the FBD can have complete control over the
  # remotes it uses.
  # If more control is needed we'll have to add it.
  accepted_remote_urls = [
     GITHUB_REPO_SSH_URL(repo_name_with_owner=searched_repo), # ssh
     f'ssh://git@github.com/{searched_repo}', # ssh
     f'https://github.com/{searched_repo}.git', # token (auth not needed for fetch)
  ]
  candidates = dict() # url->remote

  # NOTE: a shallow cloned repository has no remotes.
  for remote in repo.remotes:
    if remote.url not in accepted_remote_urls or remote.url in candidates:
      continue
    # To be honest, we'll likely encounter the (default) first refspec case
    # in almost all matching remotes.
    accepted_refspecs = {
      f'+refs/heads/*:refs/remotes/{remote.name}/*'
    }
    if branch:
      accepted_refspecs.add(
                  f'+refs/heads/{branch}:refs/remotes/{remote.name}/{branch}')
    for refspec in remote.fetch_refspecs:
      if refspec in accepted_refspecs:
        # Could ask the user here if this remote should be used
        # but actually, the most common case will be that there's just
        # one that is good, and we're picking below from the ordered list
        # of accepted_remote_urls.
        candidates[remote.url] = remote
      # else Skipping refspec is probably insufficient.

  for url in accepted_remote_urls:
    if url in candidates:
      return candidates[url].name
  return None


class PYGit2RemoteCallbacks(pygit2.RemoteCallbacks):
  # this will be set if a push was rejected
  rejected_push_message: typing.Union[str, None] = None
  def push_update_reference(self, refname, message):

    """Push update reference callback. Override with your own function to
    report the remote’s acceptance or rejection of reference updates.

    refnamestr

        The name of the reference (on the remote).
    messagestr

        Rejection message from the remote. If None, the update was accepted.
    """
    if message is not None:
      self.rejected_push_message = (f'Update to reference {refname} got '
                                    f'rejected with message {message}')

  def credentials(self, url, username_from_url, allowed_types):
    if allowed_types & pygit2.credentials.GIT_CREDENTIAL_USERNAME:
      print('GIT_CREDENTIAL_USERNAME')
      return pygit2.Username("git")
    elif allowed_types & pygit2.credentials.GIT_CREDENTIAL_SSH_KEY:
      # https://github.com/libgit2/pygit2/issues/428#issuecomment-55775298
      # "The username for connecting to GitHub over SSH is 'git'."


      # I filed https://github.com/libgit2/pygit2/issues/1013
      # because using just:
      #      return pygit2.Keypair(username_from_url, pubkey, privkey, '')
      # didn't work, there's also the example how I tried.

      # It's probably also what the user (the git command of the user)
      # does in this case and uses ssh-agent to do the auth
      #   return pygit2.Keypair(username_from_url, None, None, '')
      # There's a better readable shortcut (does the same):
      # If "git clone ..." works with an ssh remote, this should work
      # as well, no need to put configuration anywhere.
      return pygit2.KeypairFromAgent(username_from_url)
    else:
      return False
  # def sideband_progress(self, data):
  #   print(f'sideband_progress: {data}')
  #
  # # this works!
  # def transfer_progress(self, tp):
  #   print('transfer_progress:\n'
  #         f'  received_bytes {tp.received_bytes}\n'
  #         f'  indexed_objects {tp.indexed_objects}\n'
  #         f'  received_objects {tp.received_objects}')

def _git_fetch_main(repo: pygit2.Repository, remote_name: str) -> None:

  # perform a fetch
  print(f'Start fetching {remote_name}/main')
  # fetch(refspecs=None, message=None, callbacks=None, prune=0)
  # using just 'main' instead of 'refs/heads/main' works as well

  # This fails on MacOS, just as any oother pygit2 network interaction.
  # remote = repo.remotes[remote_name]
  # stats = remote.fetch(['refs/heads/main'], callbacks=PYGit2RemoteCallbacks())

  subprocess.run(['git', 'fetch', remote_name, 'main'],
    cwd=repo.path,
    check=True,
    stdout=subprocess.PIPE
  )


  print(f'DONE fetch') # {_sizeof_fmt(stats.received_bytes)} '
        # f'{stats.indexed_objects} receive dobjects!')

@contextmanager
def _create_tmp_remote(repo: pygit2.Repository, url:str) -> typing.Iterator[pygit2.Remote]:
  remote_name_template = 'tmp_{}'.format
  # create a new remote (with unique name)
  i = 0
  tmp_name: str
  remote: pygit2.Remote
  # try to create and expect to fail if it exists
  while True:
    try:
      tmp_name = remote_name_template(i)
      remote = repo.remotes.create(tmp_name, url)
      break
    except ValueError as e:
      # raises ValueError: remote '{tmp_name}' already exists
      if 'already exists' not in f'{e}':
        # ValueError is rather generic, maybe another condition can raise
        # it, hence I check for the phrase "already exists" as well.
        # I think something similar to FileExistsError would have been better
        # to raise here. Though that's an OSError.
        raise e
      i += 1
      continue
  try:
    yield remote
  finally:
    repo.remotes.delete(tmp_name)

# note: currently unused, example!
# def _git_create_remote(repo: pygit2.Repository) -> None:
#   # If we did not find a suitable remote, we can add it.
#   # If remote_name exists: repo.remotes.creat raises:
#   # "ValueError: remote 'upstream' already exists"
#   default_remote_name = 'upstream'
#
#   remote_name = input(f'Creating a git remote.\nEnter remote name (default={default_remote_name}),a=abort:')
#   if remote_name == 'a':
#     raise UserAbortError()
#   remote_name = remote_name or default_remote_name
#
#   searched_repo = 'google/fonts'
#   url =  f'git@github.com:{searched_repo}.git'
#   # url =  f'ssh://git@github.com/{searched_repo}'
#   # url = f'https://github.com/{searched_repo}.git'
#
#   refspecs_candidates = {
#       '1': f'+refs/heads/*:refs/remotes/{remote_name}/*'
#     , '2': f'+refs/heads/main:refs/remotes/{remote_name}/main'
#   }
#   print('Pick a fetch refspec for the remote:')
#   print(f'1: {refspecs_candidates["1"]} (default)')
#   print(f'2: {refspecs_candidates["2"]} (minimal)')
#   refspec = input(f'1(default),2,a=abort:').strip()
#   if refspec == 'a':
#     raise UserAbortError()
#   fetch_refspec = refspecs_candidates[refspec or '1']
#
#   # raises ValueError: remote 'upstream' already exists
#   # fetch argument will apply the default refspec if it is None
#   repo.remotes.create(remote_name, url, fetch=fetch_refspec)



#!/usr/bin/env python3
# Copyright 2016 The Fontbakery Authors
# Copyright 2017 The Google Fonts Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
import csv
import os
import sys

from fontTools import ttLib
import tabulate
from gftools.utils import get_fsSelection_byte1, get_fsSelection_byte2
from gftools.util.styles import STYLE_NAMES, is_filename_canonical
from gftools.fix import fix_fs_selection, FontFixer

parser = argparse.ArgumentParser(description='Print out fsSelection'
                                             ' bitmask of the fonts')
parser.add_argument('font', nargs="+")
parser.add_argument('--csv', default=False, action='store_true')
parser.add_argument('--usetypometrics', default=False, action='store_true')
parser.add_argument('--autofix', default=False, action='store_true')


def printInfo(fonts, print_csv=False):
  rows = []
  headers = ['filename', 'fsSelection']
  for font in fonts:
    ttfont = ttLib.TTFont(font)
    row = [os.path.basename(font)]
    row.append(('{:#010b} '
                '{:#010b}'
                '').format(get_fsSelection_byte2(ttfont),
                           get_fsSelection_byte1(ttfont)).replace('0b', ''))
    rows.append(row)

  def as_csv(rows):
    writer = csv.writer(sys.stdout)
    writer.writerows([headers])
    writer.writerows(rows)
    sys.exit(0)

  if print_csv:
    as_csv(rows)
  else:
    print(tabulate.tabulate(rows, headers, tablefmt="pipe"))

def main():
  args = parser.parse_args()
  if args.autofix:
    fixed_fonts = []
    for font in args.font:
      filename = os.path.basename(font)

      if not is_filename_canonical(filename):
        print(f"Font filename '{filename}' is not canonical!\n\n"
              f"Filename must be structured as familyname-style.ttf and "
              f"the style must be any of the following {STYLE_NAMES}")
        exit(-1)

      fixer = FontFixer(font)
      fixer.fixes = [fix_fs_selection]
      fixer.fix()
      if fixer.saveit:
        fixed_fonts.append(font)

    if len(fixed_fonts) > 0:
      printInfo([f + '.fix' for f in fixed_fonts], print_csv=args.csv)

    sys.exit(0)

  printInfo(args.font, print_csv=args.csv)


if __name__ == '__main__':
    main()




#!/usr/bin/env python3
"""Add or update a designer entry in the Google Fonts catalog.

Designer profiles are stored in the google/fonts repo:
https://github.com/google/fonts/tree/main/catalog/designers

This script is intended to help onboarders quickly create individual designer
profiles.

Designer Profiles are visible to the public. They are included in the the about
section for each family e.g:
https://fonts.google.com/specimen/Roboto#about

In 2021, Rosalie Wagner created an online form for us to collect designer
information from the general public. The form populates a private spreadsheet.

There is also an image directory where the designer's profile pics are uploaded into.

In order to use the spreadsheet command in this script, you will need to
download the folder which contains the spreadsheet and designer images.
You can then use the ``--spreadsheet`` arg.

Usage:
# Add or update a designer entry. User will need to hand complete the bio.html.
$ gftools add-designer path/to/local/clone/fonts/catalog/designers "Theo Salvadore" --img_path path/to/img.png

# Add or update a designer entry using the spreadsheet.
$ gftools add-designer path/to/local/clone/fonts/catalog/designers "Theo Salvador" --img_path path/to/img.png --spreadsheet ./GFDesigners.xlsx
"""
import argparse
from glob import glob
import os
from unidecode import unidecode
from PIL import Image
from gftools.designers_pb2 import DesignerInfoProto
from google.protobuf import text_format


def process_image(fp):
    if not os.path.isfile(fp):
        raise ValueError(f"{fp} is not a file")
    img = Image.open(fp)
    width, height = img.size
    if width != height:
        print("warning: img is rectangular when it should be square")
    if width < 300 or height < 300:
        print("warning: img is smaller than 300x300px")
        return img

    print("resizing image")
    img.thumbnail((300, 300))
    return img


def gen_info(designer, img_path=None, link=""):
    # Write info.pb
    info = DesignerInfoProto()
    info.designer = designer
    info.link = ""
    if img_path:
        info.avatar.file_name = img_path

    text_proto = text_format.MessageToString(info, as_utf8=True, use_index_order=True)
    return text_proto


def parse_urls(string):
    urls = string.split()
    res = []
    for url in urls:
        if not url.startswith("http"):
            url = "https://" + url
        res.append(url)
    return res


def gen_hrefs(urls):
    res = {}
    for url in urls:
        if "twitter" in url:
            res[url] = "Twitter"
        elif "facebook" in url:
            res[url] = "Facebook"
        elif "instagram" in url:
            res[url] = "Instagram"
        else:
            # https://www.mysite.com --> mysite.com
            res[url] = url.split("//")[1]
    return " | ".join(f'<a href="{k}">{v}</a>' for k,v in res.items())


def make_designer(
    designer_directory,
    name,
    img_path=None,
    bio=None,
    urls=None,
):
    designer_dir_name = unidecode(name.lower().replace(" ", "").replace("-", ""))
    designer_dir = os.path.join(designer_directory, designer_dir_name)
    if not os.path.isdir(designer_dir):
        print(f"{name} isn't in catalog. Creating new dir {designer_dir}")
        os.mkdir(designer_dir)

    existing_imgs = glob(os.path.join(designer_dir, "*[.png|.jpg|.jpeg]"))
    if not img_path and existing_imgs:
        img_path = existing_imgs[0]
    img_dst = os.path.join(designer_dir, f"{designer_dir_name}.png")

    if img_path and img_path in existing_imgs:
        print("Skipping image processing")
    elif img_path:
        print(f"processing image {img_path}")
        # remove any existing images
        if existing_imgs:
            [os.remove(i) for i in existing_imgs]
        image = process_image(img_path)
        image.save(img_dst)

    print(f"Generating info.pb file")
    info_pb = gen_info(name, os.path.basename(img_dst) if os.path.isfile(img_dst) else None)
    filename = os.path.join(designer_dir, "info.pb")
    with open(filename, "w") as f:
        f.write(info_pb)

    # write/update bio.html
    bio_file = os.path.join(designer_dir, "bio.html")
    html_text = None
    if bio:
        print("Generating bio.html")
        html_text = f"<p>{bio}</p>"
        if urls:
            hrefs = gen_hrefs(urls)
            html_text += "\n" + f"<p>{hrefs}</p>"
    elif os.path.isfile(bio_file):
        print("Skipping. No bio text supplied but bio.html already exists")
    else:
        print(f"Please manually update the bio.html file")
        html_text = "N/A"
    if html_text:
        with open(bio_file, "w") as f:
            f.write(html_text)
    print(f"Finished profile {designer_dir}")


def main():
    parser = argparse.ArgumentParser(usage=__doc__)
    parser.add_argument("designers_directory", help="path to google/fonts designer dir")
    parser.add_argument("name", help="Designer name e.g 'Steve Matteson'")
    parser.add_argument("--img_path", help="Optional path to profile image", default=None)
    parser.add_argument(
        "--spreadsheet", help="Optional path to the Google Drive spreadsheet"
    )
    args = parser.parse_args()

    if args.spreadsheet:
        try:
            import pandas as pd
        except ImportError as e:
            raise ValueError("The pandas library is required to read Excel spreadsheets")

        df = pd.read_excel(args.spreadsheet)
        entry = df.loc[df["Name"] == args.name]
        if len(entry) == 0:
            raise ValueError(f"Spreadsheet doesn't contain name '{args.name}'")
        bio = entry["Bio"].item()
        urls = entry["Link"].item()
        if isinstance(urls, float):  # pandas DF sets empty cells to a float
            urls = None
        else:
            urls = parse_urls(urls)

        if isinstance(bio, float):
            bio = None
    else:
        bio = None
        urls = None

    make_designer(args.designers_directory, args.name, args.img_path, bio, urls)


if __name__ == "__main__":
    main()



from gftools.builder import GFBuilder
import tempfile
import pytest
import yaml
import os
from strictyaml import load, YAMLError
from strictyaml.exceptions import YAMLValidationError


def monkey_patch_config_defaults(self):
    """we're not testing this"""
    pass


GFBuilder.fill_config_defaults = monkey_patch_config_defaults

@pytest.fixture
def config_file():
    return {
        "buildStatic": True,
        "includeSourceFixes": False,
        "instances": {
            "MavenPro[wght].ttf": [
                {"coordinates": {"wght": 500}, "familyName": "Dejon"},
                {"coordinates": {"wght": 400}, "familyName": "Dejon Special"},
                {"coordinates": {"wght": 500}, "familyName": "Bogart"},
            ]
        },
        "sources": ["MavenPro.glyphs"],
    }


def test_good_config(config_file):
    # check a good config file
    with tempfile.NamedTemporaryFile("r+", suffix=".yml") as f:
        f.write(yaml.dump(config_file))
        f.seek(0)
        GFBuilder(f.name)


def test_bad_config_structure(config_file):
    # change sources to a string when it should be a list
    config_file["sources"] = "foobar"
    with tempfile.NamedTemporaryFile("r+", suffix=".yml") as f:
        f.write(yaml.dump(config_file))
        f.seek(0)
        with pytest.raises(ValueError, match="The yaml config file isn't structured properly"):
            GFBuilder(f.name)


def test_bad_config_key(config_file):
    # Spell a key incorrectly
    config_file["source"] = config_file["sources"]
    del config_file["sources"]
    with tempfile.NamedTemporaryFile("r+", suffix=".yml") as f:
        f.write(yaml.dump(config_file))
        f.seek(0)
        with pytest.raises(YAMLError, match="A key in the configuration file"):
            GFBuilder(f.name)



"""Helper functions for creating/checking the server push files."""
import requests
from pathlib import Path
from dataclasses import dataclass
from gftools.utils import read_proto
import gftools.fonts_public_pb2 as fonts_pb2


SANDBOX_URL = "https://fonts.sandbox.google.com/metadata/fonts"
PRODUCTION_URL = "https://fonts.google.com/metadata/fonts"

PUSH_STATUS_TEMPLATE = """
***{} Status***
New families:
{}

Existing families, last pushed:
{}
"""


@dataclass
class PushItem:
    path: Path
    type: str

    def to_json(self):
        return {"path": str(self.path), "type": self.type}


def parse_server_file(fp):
    results = []
    with open(fp) as doc:
        lines = doc.read().split("\n")
        category = "Unknown"
        for line in lines:
            if not line:
                continue
            if line.startswith("#"):
                category = line[1:].strip()
            elif "#" in line:
                line = line.split("#")[0].strip()
            else:
                item = PushItem(Path(line), category)
                results.append(item)
    return results


def is_family_dir(path):
    return any(t for t in ("ofl", "apache", "ufl") if t in path.parts)


def family_dir_name(path):
    metadata_file = path / "METADATA.pb"
    assert metadata_file.exists()
    return read_proto(metadata_file, fonts_pb2.FamilyProto()).name


def gf_server_metadata(url):
    """Get family json data from a Google Fonts metadata url"""
    # can't do requests.get("url").json() since request text starts with ")]}'"
    info = requests.get(url).json()
    return {i["family"]: i for i in info["familyMetadataList"]}


def server_push_status(fp, url):
    dirs = [fp.parent / p.path for p in parse_server_file(fp)]
    family_dirs = [d for d in dirs if is_family_dir(d)]
    family_names = [family_dir_name(d) for d in family_dirs]

    gf_meta = gf_server_metadata(url)

    new_families = [f for f in family_names if f not in gf_meta]
    existing_families = [f for f in family_names if f in gf_meta]

    gf_families = sorted(
        [gf_meta[f] for f in existing_families], key=lambda k: k["lastModified"]
    )
    existing_families = [f"{f['family']}: {f['lastModified']}" for f in gf_families]
    return new_families, existing_families


def server_push_report(name, fp, server_url):
    new_families, existing_families = server_push_status(fp, server_url)
    new = "\n".join(new_families) if new_families else "N/A"
    existing = "\n".join(existing_families) if existing_families else "N/A"
    print(PUSH_STATUS_TEMPLATE.format(name, new, existing))


def push_report(fp):
    prod_path = fp / "to_production.txt"
    server_push_report("Production", prod_path, PRODUCTION_URL)

    sandbox_path = fp / "to_sandbox.txt"
    server_push_report("Sandbox", sandbox_path, SANDBOX_URL)


def missing_paths(fp):
    paths = [fp.parent / p.path for p in parse_server_file(fp)]
    axis_files = [p for p in paths if p.name.endswith("textproto")]
    dirs = [p for p in paths if p not in axis_files]

    missing_dirs = [p for p in dirs if not p.is_dir()]
    missing_axis_files = [p for p in axis_files if not p.is_file()]
    return missing_dirs + missing_axis_files


def lint_server_files(fp):
    template = "{}: Following paths are not valid:\n{}\n\n"
    prod_path = fp / "to_production.txt"
    prod_missing = "\n".join(map(str, missing_paths(prod_path)))
    prod_msg = template.format("to_production.txt", prod_missing)

    sandbox_path = fp / "to_sandbox.txt"
    sandbox_missing = "\n".join(map(str, missing_paths(sandbox_path)))
    sandbox_msg = template.format("to_sandbox.txt", sandbox_missing)

    if prod_missing and sandbox_missing:
        raise ValueError(prod_msg + sandbox_msg)
    elif prod_missing:
        raise ValueError(prod_msg)
    elif sandbox_missing:
        raise ValueError(sandbox_msg)
    else:
        print("Server files have valid paths")



#!/usr/bin/env python3
"""
gftools lang

Updates METADATA.pb to add is_noto field to families detected as Noto.

Families are determined to be part of the Noto collection based on naive logic.
Results should be verified.

Usage:

# Standard usage.
gftools tag-noto ofl/**/METADATA.pb

"""

from absl import app
from absl import flags
from gftools import fonts_public_pb2
from google.protobuf import text_format
import re


NOTO_FAMILY_NAME = re.compile(r'^Noto .*')


FLAGS = flags.FLAGS
flags.DEFINE_bool('preview', False, 'Preview mode', short_name='p')


def _ReadProto(proto, path):
  with open(path, 'r', encoding='utf-8') as f:
    proto = text_format.Parse(f.read(), proto)
    return proto


def _WriteProto(proto, path):
  with open(path, 'w', newline='') as f:
    textproto = text_format.MessageToString(proto, as_utf8=True)
    f.write(textproto)


def main(argv):
  assert len(argv) > 1, 'No METADATA.pb files specified'

  if FLAGS.preview:
    print('Running in preview mode. No changes will be made.')
    print('The names of families detected as part of the Noto')
    print('collection will be printed below.')

  for path in argv[1:]:
    family = _ReadProto(fonts_public_pb2.FamilyProto(), path)
    if NOTO_FAMILY_NAME.search(family.name):
      if FLAGS.preview:
        print(family.name)
      else:
        family.is_noto = True
        _WriteProto(family, path)


if __name__ == '__main__':
  app.run(main)



#!/usr/bin/env python3
"""
Generate the to_production.txt and to_sandbox.txt server files in a local
google/fonts repository.

to_production.txt file tells the engineers which directories need to be pushed
to the production server. Likewise, the to_sandbox.txt file is for directories
to be pushed to the sandbox server.

In order for this script to work, the traffic jam must be kept up to date and
pull requests must use labels.

Usage:
gftools gen-push-lists /path/to/google/fonts
"""
from gftools.push import parse_server_file
from github import Github
from collections import defaultdict
import os
import sys


def pr_directories(pr):
    results = set()
    files = pr.get_files()
    for f in files:
        results.add(os.path.dirname(f.filename))
    return results


def write_server_file(data):
    doc = []
    for title, directories in data.items():
        directories = sorted(directories)
        doc.append("# " + f"{title}")
        doc.append("\n".join(directories))
        doc.append("")
    return "\n".join(doc)


def main():
    if len(sys.argv) != 2:
        print("Usage: gftools gen-push-lists /path/to/google/fonts")
        sys.exit()

    to_sandbox = defaultdict(set)
    to_production = defaultdict(set)

    github = Github(os.environ["GH_TOKEN"])
    repo = github.get_repo("google/fonts")

    projects = repo.get_projects()
    traffic_jam = next((p for p in projects if p.name == "Traffic Jam"), None)
    if not traffic_jam:
        raise ValueError("Traffic Jam column has been deleted or renamed")
    columns = traffic_jam.get_columns()

    seen_directories = set()
    print("Analysing pull requests in traffic jam. This may take a while!")
    for col in columns:
        if col.name not in set(
            ["Just merged / In transit", "In Sandbox list", "In Production list"]
        ):
            continue
        cards = col.get_cards()
        for card in cards:
            content = card.get_content()
            if not hasattr(content, "labels"):
                print(f"skipping {card}. No labels!")
                continue

            labels = set(l.name for l in content.labels)
            pr = content.as_pull_request()
            directories = pr_directories(pr)
            if "-- blocked" in labels or "--- Live" in labels:
                continue
            seen_directories |= directories
            if "I Font Upgrade" in labels or "III VF Replacement" in labels:
                cat = "Upgrade"
            elif "I New Font" in labels:
                cat = "New"
            elif "I Description/Metadata/OFL" in labels:
                cat = "Metadata/desc/OFL"
            elif "I Designer profile" in labels:
                cat = "Designer profile"
            else:
                cat = "Small fix/other"
            if "--- to sandbox" in labels:
                to_sandbox[cat] |= directories
            if "--- to production" in labels:
                to_production[cat] |= directories

    gf_repo_path = sys.argv[1]
    sb_path = os.path.join(gf_repo_path, "to_sandbox.txt")
    prod_path = os.path.join(gf_repo_path, "to_production.txt")

    # Keep paths which have been entered manually which do not belong to
    # a label. These need to be manually deleted as well.
    existing_sandbox = parse_server_file(sb_path)
    for i in existing_sandbox:
        if str(i.path) not in seen_directories:
            to_sandbox[i.type].add(str(i.path))

    existing_production = parse_server_file(prod_path)
    for i in existing_production:
        if str(i.path) not in seen_directories:
            to_production[i.type].add(str(i.path))

    with open(sb_path, "w") as sb_doc:
        sb_doc.write(write_server_file(to_sandbox))

    with open(prod_path, "w") as prod_doc:
        prod_doc.write(write_server_file(to_production))

    print("Done!")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2020 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
from gftools.builder import GFBuilder
from gftools.builder import __doc__ as GFBuilder_doc


parser = argparse.ArgumentParser(
    description=("Build a font family"),
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog="#"*79 + "\n" + GFBuilder_doc,
)
parser.add_argument(
    "--debug",
    action="store_true",
    default=False,
    help="Show extra debugging information",
)
parser.add_argument("--family-name", help="Font family name")
parser.add_argument(
    "--no-autohint",
    action="store_true",
    default=False,
    help="Don't run ttfautohint on static TTFs",
)
parser.add_argument("--stylespace", help="Path to a statmake stylespace file")

parser.add_argument(
    "--no-clean-up",
    action="store_true",
    default=False,
    help="Do not remove temporary files (instance_ufos/)")

parser.add_argument("file", nargs="+", help="YAML build config file *or* source files")

parser.add_argument("--dump-config", type=str, help="Config file to generate")

args = parser.parse_args()

if len(args.file) == 1 and (
    args.file[0].endswith(".yaml") or args.file[0].endswith(".yml")
):
    builder = GFBuilder(configfile=args.file[0])
else:
    config={"sources": args.file}
    if args.stylespace:
        config["stylespaceFile"] = args.stylespace
    if args.family_name:
        config["familyName"] = args.family_name
    builder = GFBuilder(config=config)

if args.no_autohint:
    builder.config["autohintTTF"] = False

if args.no_clean_up:
    builder.config["cleanUp"] = False

if args.debug:
    builder.config["logLevel"] = "DEBUG"

if args.dump_config:
    import sys
    import yaml

    with open(args.dump_config, "w") as fp:
        config= {k: v for (k, v) in builder.config.items() if v is not None}
        fp.write(yaml.dump(config, Dumper=yaml.SafeDumper))
    sys.exit()

builder.build()



import os
import pprint
import requests
import typing
import urllib


GITHUB_GRAPHQL_API = 'https://api.github.com/graphql'
GITHUB_V3_REST_API = 'https://api.github.com'


class GitHubClient:
    def __init__(self, repo_owner, repo_name):
        if not 'GH_TOKEN' in os.environ:
            raise Exception("GH_TOKEN environment variable not set")
        self.gh_token = os.environ['GH_TOKEN']
        self.repo_owner = repo_owner
        self.repo_name = repo_name
    
    def _post(self, url, payload: typing.Dict):
        headers = {'Authorization': f'bearer {self.gh_token}'}
        response = requests.post(url, json=payload, headers=headers)
        if response.status_code == requests.codes.unprocessable:
            # has a helpful response.json with an 'errors' key.
            pass
        else:
            response.raise_for_status()
        json = response.json()
        if 'errors' in json:
            errors = pprint.pformat(json['errors'], indent=2)
            raise Exception(f'GitHub POST query failed to url {url}:\n {errors}')
        return json
    
    def _get(self, url):
        headers = {'Authorization': f'bearer {self.gh_token}'}
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        json = response.json()
        if 'errors' in json:
            errors = pprint.pformat(json['errors'], indent=2)
            raise Exception(f'GitHub REST query failed:\n {errors}')
        return json
        
    def _run_graphql(self, query, variables):
        payload = {'query': query, 'variables': variables}
        return self._post(GITHUB_GRAPHQL_API, payload)
    
    def rest_url(self, path, **kwargs):
        base_url = f'{GITHUB_V3_REST_API}/repos/{self.repo_owner}/{self.repo_name}/{path}'
        if kwargs:
            base_url += '?' + '&'.join(f'{k}={urllib.parse.quote(v)}' for k, v in kwargs.items())
        return base_url

    def get_blob(self, file_sha):
        url = self.rest_url(f'git/blobs/{file_sha}')
        headers = {
            'Accept': 'application/vnd.github.v3.raw',
            'Authorization': f'bearer {self.gh_token}'
        }
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response
    
    def open_prs(self, pr_head: str, pr_base_branch: str) -> typing.List:
        return self._get(self.rest_url("pulls", state="open", head=pr_head, base=pr_base_branch))
    
    def create_pr(self, title: str, body: str, head: str, base: str):
        return self._post(
            self.rest_url("pulls"),
            {
                'title': title,
                'body': body,
                'head': head,
                'base': base,
                'maintainer_can_modify': True
            }
        )
    
    def create_issue_comment(self, issue_number: int, body: str):
        return self._post(
            self.rest_url(f'issues/{issue_number}/comments'),
            {
                'body': body
            }
        )
    
    def create_issue(self, title: str, body: str):
        return self._post(
            self.rest_url("issues"),
            {
                'title': title,
                'body': body
            }
        )









#!/usr/bin/env python3
"""
gftools fix-font

Update a font so it conforms to the Google Fonts specification
https://github.com/googlefonts/gf-docs/tree/main/Spec

Usage:

gftools fix-font font.ttf

# Fix font issues that should be fixed in the source files
gftools fix-font font.ttf --include-source-fixes
"""
import argparse
import logging
from fontTools.ttLib import TTFont
from gftools.fix import *


log = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("font", help="Path to font")
    parser.add_argument("-o", "--out", help="Output path for fixed font")
    parser.add_argument(
        "--include-source-fixes",
        action="store_true",
        help="Fix font issues that should be fixed in the source files.",
    )
    args = parser.parse_args()

    font = TTFont(args.font)

    fix_font(font, args.include_source_fixes)

    if args.out:
        font.save(args.out)
    else:
        font.save(font.reader.file.name + ".fix")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility to dump codepoints in a font.

Prints codepoints supported by the font, one per line, in hex (0xXXXX).

"""

import os
import sys
import unicodedata
from absl import (app,
                  flags)
from glyphsets.codepoints import (CodepointsInFont,
                                  SubsetsForCodepoint)

if sys.version[0] == '3':
    unichr = chr

FLAGS = flags.FLAGS
flags.DEFINE_bool('show_char', False, 'Print the actual character')
flags.DEFINE_bool('show_subsets', False,
                  'Print what subsets, if any, char is in')


def main(argv):
  if len(argv) < 2:
    sys.exit('Must specify one or more font files.')

  cps = set()
  for filename in argv[1:]:
    if not os.path.isfile(filename):
      sys.exit('%s is not a file' % filename)
    cps |= CodepointsInFont(filename)

  for cp in sorted(cps):
    show_char = ''
    if FLAGS.show_char:
      show_char = (' ' + unichr(cp).strip() + ' ' +
                   unicodedata.name(unichr(cp), ''))
    show_subset = ''
    if FLAGS.show_subsets:
      show_subset = ' subset:%s' % ','.join(SubsetsForCodepoint(cp))

    print(u'0x%04X%s%s' % (cp, show_char, show_subset))

if __name__ == '__main__':
  app.run(main)



#!/usr/bin/env python3
# Copyright 2013 The Font Bakery Authors.
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See AUTHORS.txt for the list of Authors and LICENSE.txt for the License.
#
import argparse
from fontTools import ttLib
from gftools.fix import fix_ascii_fontmetadata


description = 'Fixes TTF NAME table strings to be ascii only'
parser = argparse.ArgumentParser(description=description)
parser.add_argument('ttf_font', nargs='+',
                    help="Font in OpenType (TTF/OTF) format")

def main():
  args = parser.parse_args()
  for path in args.ttf_font:
      fix_ascii_fontmetadata(ttLib.TTFont(path))

if __name__ == "__main__":
  main()



#!/usr/bin/env python3
# Copyright 2016 The Fontbakery Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import requests
from io import BytesIO
from zipfile import ZipFile
import sys
import os
import shutil
import unicodedata
from unidecode import unidecode
from collections import namedtuple
from github import Github
from pkg_resources import resource_filename
from google.protobuf import text_format
import json
from PIL import Image
if sys.version_info[0] == 3:
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser

# =====================================
# HELPER FUNCTIONS

def download_family_from_Google_Fonts(family, dst=None):
    """Download a font family from Google Fonts"""
    url = 'https://fonts.google.com/download?family={}'.format(
        family.replace(' ', '%20')
    )
    fonts_zip = ZipFile(download_file(url))
    if dst:
        fonts = fonts_from_zip(fonts_zip, dst)
        # Remove static fonts if the family is a variable font
        return [f for f in fonts if "static" not in f]
    return fonts_from_zip(fonts_zip)


def Google_Fonts_has_family(name):
    """Check if Google Fonts has the specified font family"""
    # This endpoint is private and may change at some point
    # TODO (MF) if another function needs this data, refactor it into a
    # function and use a lru cache
    data = requests.get("https://fonts.google.com/metadata/fonts").json()
    family_names = set(i["family"] for i in data["familyMetadataList"])
    return name in family_names


def load_Google_Fonts_api_key():
    config = ConfigParser()
    config_filepath = os.path.expanduser("~/.gf-api-key")

    if os.path.isfile(config_filepath):
        config.read(config_filepath)
        credentials = config.items("Credentials")
        return credentials[0][1]
    return None


def parse_github_pr_url(url):
    if not "github.com" in url and "pull" not in url:
        raise ValueError("{} is not a github.com pr url".format(url))
    if not url[-1].isdigit():
        raise ValueError("{} should end with a pull request number".format(url))
    segments = url.split("/")
    GithubPR = namedtuple("GithubPR", "user repo pull")
    return GithubPR(segments[3], segments[4], int(segments[-1]))


def parse_github_dir_url(url):
    if not "github.com" in url:
        raise ValueError("{} is not a github.com dir url".format(url))
    segments = url.split("/")
    GithubDir = namedtuple("GithubDir", "user repo branch dir")
    return GithubDir(segments[3], segments[4], segments[6], "/".join(segments[7:]))


def download_files_in_github_pr(
    url,
    dst,
    filter_files=[],
    ignore_static_dir=True,
    overwrite=True,
):
    """Download files in a github pr e.g
    https://github.com/google/fonts/pull/2072

    Arguments
    ---------
    url: str, github pr url
    dst: str, path to output files
    filter_files: list, collection of files to include. None will keep all.
    ignore_static_dir: bool, If true, do not include files which reside in
    a /static dir. These dirs are used in family dirs on google/fonts
    e.g ofl/oswald.
    overwrite: bool, set True to overwrite existing contents in dst

    Returns
    -------
    list of paths to downloaded files
    """
    gh = Github(os.environ["GH_TOKEN"])
    url = parse_github_pr_url(url)
    repo_slug = "{}/{}".format(url.user, url.repo)
    repo = gh.get_repo(repo_slug)
    pull = repo.get_pull(url.pull)
    files = [f for f in pull.get_files()]

    mkdir(dst, overwrite=overwrite)
    # if the pr is from google/fonts or a fork of it, download all the
    # files inside the family dir as well. This way means we can qa
    # the whole family together as a whole unit. It will also download
    # the metadata, license and description files so all Fontbakery
    # checks will be executed.
    if pull.base.repo.name == "fonts":
        dirs = set([os.path.dirname(p.filename) for p in files])
        results = []
        for d in dirs:
            if ignore_static_dir and '/static' in d:
                continue
            url = os.path.join(
                pull.head.repo.html_url,
                "tree",
                pull.head.ref, # head branch
                d)
            results += download_files_in_github_dir(url, dst, overwrite=False)
        return results

    results = []
    for f in files:
        filename = os.path.join(dst, f.filename)
        if filter_files and not filename.endswith(tuple(filter_files)):
            continue
        if ignore_static_dir and "/static" in filename:
            continue
        if not overwrite and os.path.exists(filename):
            continue
        dst_ = os.path.dirname(filename)
        mkdir(dst_, overwrite=False)
        download_file(f.raw_url, filename)
        results.append(filename)
    return results


def download_files_in_github_dir(
    url,
    dst,
    filter_files=[],
    overwrite=True
):
    """Download files in a github dir e.g
    https://github.com/google/fonts/tree/main/ofl/abhayalibre

    Arguments
    ---------
    url: str, github dir url
    dst: str, path to output files
    filter_files: list, collection of files to include. None will keep all.
    overwrite: bool, set True to overwrite existing contents in dst

    Returns
    -------
    list of paths to downloaded files
    """
    gh = Github(os.environ["GH_TOKEN"])
    url = parse_github_dir_url(url)
    repo_slug = "{}/{}".format(url.user, url.repo)
    repo = gh.get_repo(repo_slug)
    files = [f for f in repo.get_contents(url.dir, ref=url.branch)
             if f.type == 'file']

    mkdir(dst, overwrite=overwrite)
    results = []
    for f in files:
        filename = os.path.join(dst, f.path)
        if filter_files and not filename.endswith(tuple(filter_files)):
            continue
        if not overwrite and os.path.exists(filename):
            continue
        dst_ = os.path.dirname(filename)
        mkdir(dst_, overwrite=False)
        download_file(f.download_url, filename)
        results.append(filename)
    return results


def download_file(url, dst_path=None):
    """Download a file from a url. If no dst_path is specified, store the file
    as a BytesIO object"""
    request = requests.get(url, stream=True)
    if not dst_path:
        return BytesIO(request.content)
    with open(dst_path, 'wb') as downloaded_file:
        downloaded_file.write(request.content)


def fonts_from_zip(zipfile, dst=None):
    """Unzip fonts. If not dst is given unzip as BytesIO objects"""
    fonts = []
    for filename in zipfile.namelist():
        if filename.endswith(".ttf"):
            if dst:
                target = os.path.join(dst, filename)
                zipfile.extract(filename, dst)
                fonts.append(target)
            else:
                fonts.append(BytesIO(zipfile.read(filename)))
    return fonts


def cmp(x, y):
    """
    Replacement for built-in function cmp that was removed in Python 3

    Compare the two objects x and y and return an integer according to
    the outcome. The return value is negative if x < y, zero if x == y
    and strictly positive if x > y.
    """

    return (x > y) - (x < y)


def mkdir(path, overwrite=True):
    if os.path.isdir(path) and overwrite:
        shutil.rmtree(path)
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


## Font-related utility functions

def font_stylename(ttFont):
    """Get a font's stylename using the name table. Since our fonts use the
    RIBBI naming model, use the Typographic SubFamily Name (NAmeID 17) if it
    exists, otherwise use the SubFamily Name (NameID 2).

    Args:
        ttFont: a TTFont instance
    """
    return get_name_record(ttFont, 17, fallbackID=2)


def font_familyname(ttFont):
    """Get a font's familyname using the name table. since our fonts use the
    RIBBI naming model, use the Typographic Family Name (NameID 16) if it
    exists, otherwise use the Family Name (Name ID 1).

    Args:
        ttFont: a TTFont instance
    """
    return get_name_record(ttFont, 16, fallbackID=1)


def get_name_record(ttFont, nameID, fallbackID=None, platform=(3, 1, 0x409)):
    """Return a name table record which has the specified nameID.

    Args:
        ttFont: a TTFont instance
        nameID: nameID of name record to return,
        fallbackID: if nameID doesn't exist, use this nameID instead
        platform: Platform of name record. Default is Win US English

    Returns:
        str
    """
    name = ttFont["name"]
    record = name.getName(nameID, 3, 1, 0x409)
    if not record and fallbackID:
        record = name.getName(fallbackID, 3, 1, 0x409)
    if not record:
        raise ValueError(f"Cannot find record with nameID {nameID}")
    return record.toUnicode()


def family_bounding_box(ttFonts):
    y_min = min(f["head"].yMin for f in ttFonts)
    y_max = max(f["head"].yMax for f in ttFonts)
    return y_min, y_max


def typo_metrics_enabled(ttFont):
    return True if ttFont["OS/2"].fsSelection & (1 << 7) else False


def family_is_vf(ttFonts):
    has_fvar = ["fvar" in ttFont for ttFont in ttFonts]
    if any(has_fvar):
        if all(has_fvar):
            return True
        raise ValueError("Families cannot contain both static and variable fonts")
    return False


def validate_family(ttFonts):
    family_is_vf(ttFonts)
    family_names = set(font_familyname(f) for f in ttFonts)
    if len(family_names) != 1:
        raise ValueError(f"Multiple families found {family_names}")
    return True


def unique_name(ttFont, nameids):
    font_version = _font_version(ttFont)
    vendor = ttFont["OS/2"].achVendID.strip()
    ps_name = nameids[6]
    return f"{font_version};{vendor};{ps_name}"


def _font_version(font, platEncLang=(3, 1, 0x409)):
    nameRecord = font["name"].getName(5, *platEncLang)
    if nameRecord is None:
        return f'{font["head"].fontRevision:.3f}'
    # "Version 1.101; ttfautohint (v1.8.1.43-b0c9)" --> "1.101"
    # Also works fine with inputs "Version 1.101" or "1.101" etc
    versionNumber = nameRecord.toUnicode().split(";")[0]
    return versionNumber.lstrip("Version ").strip()


def partition_cmap(font, test, report=True):
  """Drops all cmap tables from the font which do not pass the supplied test.

  Arguments:
    font: A ``TTFont`` instance
    test: A function which takes a cmap table and returns True if it should
      be kept or False if it should be removed from the font.
    report: Reports to stdout which tables were dropped and which were kept.

  Returns two lists: a list of `fontTools.ttLib.tables._c_m_a_p.*` objects
  which were kept in the font, and a list of those which were removed."""
  keep = []
  drop = []

  for index, table in enumerate(font['cmap'].tables):
    if test(table):
      keep.append(table)
    else:
      drop.append(table)

  if report:
    for table in keep:
        print(("Keeping format {} cmap subtable with Platform ID = {}"
               " and Encoding ID = {}").format(table.format,
                                               table.platformID,
                                               table.platEncID))
    for table in drop:
        print(("--- Removed format {} cmap subtable with Platform ID = {}"
             " and Encoding ID = {} ---").format(table.format,
                                                 table.platformID,
                                                 table.platEncID))

  font['cmap'].tables = keep
  return keep, drop


def _unicode_marks(string):
    unicodemap = [(u'©', '(c)'), (u'®', '(r)'), (u'™', '(tm)')]
    return filter(lambda char: char[0] in string, unicodemap)


def normalize_unicode_marks(string):
    """ Converts special characters like copyright,
        trademark signs to ascii name """
    # print("input: '{}'".format(string))
    input_string = string
    for mark, ascii_repl in _unicode_marks(string):
        string = string.replace(mark, ascii_repl)

    rv = []
#    for c in unicodedata.normalize('NFKC', smart_text(string)):
    for c in unicodedata.normalize('NFKC', string):
        # cat = unicodedata.category(c)[0]
        # if cat in 'LN' or c in ok:
        rv.append(c)

    new = ''.join(rv).strip()
    result = unidecode(new)
    if result != input_string:
        print("Fixed string: '{}'".format(result))
    return result


def get_fsSelection_byte2(ttfont):
    return ttfont['OS/2'].fsSelection >> 8


def get_fsSelection_byte1(ttfont):
    return ttfont['OS/2'].fsSelection & 255


def get_encoded_glyphs(ttFont):
    """Collect all encoded glyphs"""
    return list(map(chr, ttFont.getBestCmap().keys()))


def get_unencoded_glyphs(font):
    """ Check if font has unencoded glyphs """
    cmap = font['cmap']

    new_cmap = cmap.getcmap(3, 10)
    if not new_cmap:
        for ucs2cmapid in ((3, 1), (0, 3), (3, 0)):
            new_cmap = cmap.getcmap(ucs2cmapid[0], ucs2cmapid[1])
            if new_cmap:
                break

    if not new_cmap:
        return []

    diff = list(set(font.getGlyphOrder()) -
                set(new_cmap.cmap.values()) - {'.notdef'})
    return [g for g in diff[:] if g != '.notdef']


def has_mac_names(ttfont):
    """Check if a font has Mac names. Mac names have the following
    field values:
    platformID: 1, encodingID: 0, LanguageID: 0"""
    for i in range(255):
        if ttfont['name'].getName(i, 1, 0, 0):
            return True
    return False


def font_is_italic(ttfont):
    """Check if the font has the word "Italic" in its stylename."""
    stylename = ttfont["name"].getName(2, 3, 1, 0x409).toUnicode()
    return True if "Italic" in stylename else False


def font_sample_text(ttFont):
    """Collect words which exist in the Universal Declaration of Human Rights
    that can be formed using the ttFont instance.

    UDHR has been chosen due to the many languages it covers"""
    with open(resource_filename("gftools", "udhr_all.txt")) as doc:
        text = doc.read()

    cmap = set(ttFont.getBestCmap())
    words = []
    seen_chars = set()
    for word in text.split():
        chars = set(ord(l) for l in word)
        if not chars.issubset(cmap):
            continue
        if chars & seen_chars == chars:
            continue
        seen_chars |= chars
        words.append(word)
    return words


def gen_gifs(dir1, dir2, dst_dir):
    dir1_imgs = set(f for f in os.listdir(dir1) if f.endswith(("jpg", "png")))
    dir2_imgs = set(f for f in os.listdir(dir2) if f.endswith(("jpg", "png")))
    shared_imgs = dir1_imgs & dir2_imgs
    for img in shared_imgs:
        gif_filename = img[:-4] + '.gif'
        img_a_path = os.path.join(dir1, img)
        img_b_path = os.path.join(dir2, img)
        dst = os.path.join(dst_dir, gif_filename)
        gen_gif(img_a_path, img_b_path, dst)


def gen_gif(img_a_path, img_b_path, dst):
    with Image.open(img_a_path) as img_a, Image.open(img_b_path) as img_b:
        img_a.save(
            dst, 
            save_all=True,
            append_images=[img_b],
            loop=10000,
            duration=1000
        )


def partition(items, size):
    """partition([1,2,3,4,5,6], 2) --> [[1,2],[3,4],[5,6]]"""
    return [items[i : i + size] for i in range(0, len(items), size)]


def read_proto(fp, schema):
    with open(fp, "rb") as f:
        data = text_format.Parse(f.read(), schema)
    return data




#!/usr/bin/env python3
# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Check the status of families being pushed to Google Fonts.

Families are pushed to a sandbox server and inspected before
they are sent to the production server. The files "to_sandbox.txt" and
"to_production.txt" in the google/fonts repo list which families need
pushing to their respective servers.

This script will check whether the families listed in the text files
have been pushed. A lint command is also provided to ensure they list
valid directory paths.

Usage:
gftools push-status /path/to/google/fonts/repo
# Check server push files are valid
gftools push-status /path/to/google/fonts/repo --lint
"""
import argparse
from pathlib import Path
from gftools.push import lint_server_files, push_report


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=Path, help="Path to google/fonts repo")
    parser.add_argument(
        "--lint", action="store_true", help="Check server files have correct paths"
    )
    args = parser.parse_args()
    if args.lint:
        lint_server_files(args.path)
    else:
        push_report(args.path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Extracts text samples from UDHR translations.

Translations exist for myriad languages at varying levels of support. The tool
focuses on Stage 4+ translations for which the UDHR translation has a reliable
amount of content and structure for scraping text samples.

See more at https://www.unicode.org/udhr.
"""

import enum
import os
import ssl
import tempfile
from urllib import request
from lxml import etree
import zipfile

UHDR_URL_TEMPLATE = 'https://www.un.org/{language_code}/universal-declaration-human-rights/'
UDHR_TRANSLATIONS_ZIP_URL = 'https://www.unicode.org/udhr/assemblies/udhr_xml.zip'
INDEX_XML = 'index.xml'


class UdhrTranslations():

  def __init__(self):
    self._zip_dir = tempfile.TemporaryDirectory()
    self._DownloadUdhrTranslationsZip()
    self._udhrs = self._ParseUdhrs()
    self._udhr_map = {}

    for udhr in self._udhrs:
      udhr.Parse(self._LoadUdhrTranslation(udhr))
      self._udhr_map[udhr.key] = udhr
      self._udhr_map[udhr.iso639_3] = udhr
      self._udhr_map[udhr.iso15924] = udhr
      self._udhr_map[udhr.bcp47] = udhr

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self._zip_dir.cleanup()

  def _DownloadUdhrTranslationsZip(self):
    with tempfile.NamedTemporaryFile(suffix='.zip') as zip_file:
      # Disable SSL verification
      ssl._create_default_https_context = ssl._create_unverified_context
      request.urlretrieve(UDHR_TRANSLATIONS_ZIP_URL, zip_file.name)
      with zipfile.ZipFile(zip_file.name, 'r') as zip_ref:
        zip_ref.extractall(self._zip_dir.name)

  def _ParseUdhrs(self):
    root = etree.parse(os.path.join(self._zip_dir.name, INDEX_XML))
    return [self.Udhr(udhr_data, self._zip_dir) for udhr_data in root.xpath('*')]

  def _LoadUdhrTranslation(self, udhr):
    filename = 'udhr_{key}.xml'.format(key=udhr.key)
    path = os.path.join(self._zip_dir.name, filename)
    if os.path.exists(path):
      return etree.parse(path)
    return None

  def GetUdhrs(self, min_stage=0):
    return [udhr for udhr in self._udhrs if udhr.stage >= min_stage]

  def GetUdhr(self, lang_code, min_stage=0):
    if lang_code not in self._udhr_map or self._udhr_map[lang_code].stage < min_stage:
      return None
    return self._udhr_map[lang_code]

  class Udhr():

    def __init__(self, udhr_data, zip_dir):
      self.key = udhr_data.get('f')
      self.iso639_3 = udhr_data.get('iso639-3')
      self.iso15924 = udhr_data.get('iso15924')
      self.bcp47 = udhr_data.get('bcp47')
      self.direction = udhr_data.get('dir')
      self.ohchr = udhr_data.get('ohchr')
      self.stage = int(udhr_data.get('stage'))
      self.loc = udhr_data.get('loc')
      self.name = udhr_data.get('n')

    def Parse(self, translation_data):
      if translation_data is None or self.stage < 2:
        return

      self.title = None
      if translation_data.find('./{*}title') is not None:
        self.title = translation_data.find('./{*}title').text

      preamble_data = translation_data.find('./{*}preamble')
      self.preamble = None
      if preamble_data is not None:
        if preamble_data.find('./{*}title') is not None:
          self.preamble = {
              'title':
                  preamble_data.find('./{*}title').text,
              'content': [
                  para.text for para in preamble_data.findall('./{*}para')
                      ],
          }

      articles_data = translation_data.findall('./{*}article')
      self.articles = []
      for article_data in articles_data:
        article = {
            'id':
                int(article_data.get('number')),
            'title':
                article_data.find('./{*}title').text,
            'content': [
                para.text for para in article_data.findall('./{*}para')
                    ],
        }
        self.articles.append(article)

    def GetSampleTexts(self):
      extractor = SampleTextExtractor(udhr)
      return extractor.GetSampleTexts()

  class TextType(enum.Enum):
    GLYPHS = 1
    WORD = 2
    PHRASE = 3
    SENTENCE = 4
    PARAGRAPH = 5
    PASSAGE = 6

  class Size(enum.Enum):
    SMALL = 1
    MEDIUM = 2
    LARGE = 3

    def Switch(options, size):
      if size not in options:
        raise Error('Size {size} not in options: {options}'.format(
            options=options, size=size))
      return options[size]

  class SampleTextExtractor():

    def __init__(self, udhr):
      self._udhr = udhr
      self._glyphs = {}
      self._words = {}
      self._phrases = {}
      self._sentences = {}
      self._paragraphs = {}
      self._passages = {}

    def _ExtractGlyphs(self, size):
      if size == Size.SMALL:
        # single glyph
        pass
      elif size == Size.MEDIUM:
        # upper and lower of single glyph
        pass
      elif size == Size.LARGE:
        # upper and lower of two glyphs
        pass
      else:
        self._UnsupportedSize(size)
      pass

    def _ExtractWord(self, size):
      options = {
          Size.SMALL: (3, 5),
          Size.MEDIUM: (6, 8),
          Size.LARGE: (8, 11),
      }
      min_length, max_length = Size.Switch(options, size)
      pass

    def _ExtractPhrase(self, size):
      options = {
          Size.SMALL: (3, 7),
          Size.MEDIUM: (8, 14),
          Size.LARGE: (12, 20),
      }
      min_length, max_length = Size.Switch(options, size)
      pass

    def _ExtractSentence(self, size):
      options = {
          Size.SMALL: (6, 10),
          Size.MEDIUM: (12, 20),
          Size.LARGE: (23, 35),
      }
      min_length, max_length = Size.Switch(options, size)
      pass

    def _ExtractParagraph(self, size):
      options = {
          Size.SMALL: (10, 20),
          Size.MEDIUM: (30, 50),
          Size.LARGE: (70, 100),
      }
      length = Size.Switch(options, size)
      pass

    def _ExtractPassage(self, size):
      options = {
          Size.SMALL: 2,
          Size.MEDIUM: 3,
          Size.LARGE: 5,
      }
      length = Size.Switch(options, size)
      pass

    def _UnsupportedSize(self, size):
      raise Error('Unsupported size: ' + size)

    def _Get(self, text_type, size):
      pass

    def GetSampleTexts(self):
      return {
          'hero': [
              self._Get(TextType.GLYPHS, Size.LARGE),  # Single glyphs (eg AbZx)
          ],
          'type_tester': [
              self._Get(TextType.PHRASE, Size.LARGE),  # 12-20 word phrase
          ],
          'poster': [
              self._Get(TextType.WORD, Size.MEDIUM),  # Single word
              self._Get(TextType.WORD, Size.LARGE),  # Single longer word
              self._Get(TextType.SENTENCE, Size.SMALL),  # 6-10 word sentence
              self._Get(TextType.SENTENCE, Size.MEDIUM),  # 12-20 word sentence
          ],
          'specimen': [
              self._Get(TextType.SENTENCE, Size.SMALL),  # 6-10 word sentence
              self._Get(TextType.PARAGRAPH, Size.SMALL),  # Paragraph
              self._Get(TextType.PARAGRAPH, Size.MEDIUM),  # Medium paragraph
              self._Get(TextType.PARAGRAPH, Size.LARGE),  # Large paragraph
              self._Get(TextType.PASSAGE,
                        Size.SMALL),  # Multiple paragraphs split over 2 columns
              self._Get(TextType.PASSAGE,
                        Size.MEDIUM),  # Multiple paragraphs split over 3 columns
          ],
      }
